# -*- coding: utf-8 -*-

import warnings

try:
    from numpy import any as _any
except ImportError:
    _any = any


def water_density(T=None, T0=None, units=None, a=None,
                  just_return_a=False, warn=True):
    """
    Density of water (kg/m3) as function of temperature (K)
    according to VSMOW model between 0 and 40 degree Celsius.
    Fitted using Thiesen's equation.

    Parameters
    ----------
    T: float
        Temperature (in Kelvin) (default: 298.15)
    T0: float
        Value of T for 0 degree Celsius (default: 273.15)
    units: object (optional)
        object with attributes: Kelvin, meter, kilogram
    a: array_like (optional)
        5 parameters to the equation.
    just_return_a: bool (optional, default: False)
        Do not compute rho, just return the parameters ``a``.
    warn: bool (default: True)
        Emit UserWarning when outside temperature range.

    Returns
    -------
    Density of water (float of kg/m3 if T is float and units is None)

    Examples
    --------
    >>> print('%.2f' % water_density(277.13))
    999.97

    References
    ----------
    TANAKA M., GIRARD G., DAVIS R., PEUTO A. and BIGNELL N.,
        "Recommanded table for the density of water between 0 °C and 40 °C
        based on recent experimental reports",
        Metrologia, 2001, 38, 301-309.
        http://iopscience.iop.org/article/10.1088/0026-1394/38/4/3
        doi:10.1088/0026-1394/38/4/3
    """
    if units is None:
        K = 1
        m = 1
        kg = 1
    else:
        K = units.Kelvin
        m = units.meter
        kg = units.kilogram
    if T is None:
        T = 298.15*K
    m3 = m**3
    if a is None:
        a = (-3.983035*K,  # C
             301.797*K,  # C
             522528.9*K*K,  # C**2
             69.34881*K,  # C
             999.974950*kg/m3)
    if just_return_a:
        return a
    if T0 is None:
        T0 = 273.15*K
    t = T - T0
    if warn and (_any(t < 0*K) or _any(t > 40*K)):
        warnings.warn("Temperature is outside range (0-40 degC)")
    return a[4]*(1-((t + a[0])**2*(t + a[1]))/(a[2]*(t + a[3])))

# generated at doi2bib.org:
bibtex = """
@article{Tanaka2001,
  doi = {10.1088/0026-1394/38/4/3},
  url = {http://dx.doi.org/10.1088/0026-1394/38/4/3},
  year  = {2001},
  month = {aug},
  publisher = {{IOP} Publishing},
  volume = {38},
  number = {4},
  pages = {301--309},
  author = {M Tanaka and G Girard and R Davis and A Peuto and N Bignell},
  title = {Recommended table for the density of water between 0 ~C and 40
           ~C based on recent experimental reports},
  journal = {Metrologia}
}
"""



# -*- coding: utf-8 -*-
"""
Utility functions related to stoichiometry.
"""

from __future__ import (absolute_import, division, print_function)

import numpy as np


def get_coeff_mtx(substances, stoichs):
    """
    Create a net stoichiometry matrix from reactions
    described by pairs of dictionaries.

    Parameters
    ----------
    substances: sequence of keys in stoichs dict pairs
    stoichs: sequence of pairs of dicts
        pairs of reactant and product dicts mapping substance keys
        to stoichiometric coefficients (integers)

    Returns
    -------
    2 dimensional array of shape (len(substances), len(stoichs))

    """
    A = np.zeros((len(substances), len(stoichs)), dtype=int)
    for ri, sb in enumerate(substances):
        for ci, (reac, prod) in enumerate(stoichs):
            A[ri, ci] = prod.get(sb, 0) - reac.get(sb, 0)
    return A


def decompose_yields(yields, stoichs, atol=1e-10):
    """ Decomposes yields into mass-action reactions

    This function offers a way to express a reaction with non-integer
    stoichiometric coefficients as a linear combination of production reactions
    with integer coefficients.

    Ak = y

    A is (n_species x n_reactions) matrix, k is "rate coefficient", y is yields


    Parameters
    ----------
    yields: OrderedDict
        specie names as keys and yields as values
    stoichs: iterable dictionary pairs
        dict keys must match those of `yields`each pair
        of dictionaries gives stoichiometry
        (1st is reactant, 2nd is products)
    atol: float
        absolute tolerance for residuals

    Raises
    ------
    ValueError
        When atol is exceeded
    numpy.LinAlgError
        When numpy.linalg.lstsq fails to converge

    Returns
    -------
    1-dimensional array of effective rate coefficients.

    """
    # Sanity check:
    for ys in yields.keys():
        present = False
        for reac, prod in stoichs:
            if ys in reac or ys in prod:
                present = True
        assert present

    sbstncs = yields.keys()
    y = np.array(list(yields.values()))
    A = get_coeff_mtx(sbstncs, stoichs)
    k, residuals, rank, s = np.linalg.lstsq(A, y)
    if len(residuals) > 0:
        if np.any(residuals > atol):
            raise ValueError("atol not satisfied")
    return k



# -*- coding: utf-8 -*-

import os
import shutil
import tempfile
from chempy.chemistry import Reaction, ReactionSystem, Substance
from chempy.util.graph import rsys2dot, rsys2graph


def _get_rsys():
    r1 = Reaction({'A': 2}, {'B': 1}, param=3.0)
    A = Substance('A', latex_name='\\boldsymbol{A}')
    B = Substance('B', latex_name='\\boldsymbol{B}')
    rsys = ReactionSystem([r1], [A, B])
    return rsys


def test_rsys2dot():
    rsys = _get_rsys()
    assert list(map(str.strip, rsys2dot(rsys))) == [
        'digraph None{',
        '{',
        'node [label=r1 shape=diamond]',
        'r1',
        '}',
        '"A" -> "r1" [label ="2"];',
        '"r1" -> "B" [label =""];',
        '}'
    ]


def test_rsys2graph():
    rsys = _get_rsys()
    tempdir = tempfile.mkdtemp()
    try:
        rsys2graph(rsys, os.path.join(tempdir, 'out.png'))
        rsys2graph(rsys, os.path.join(tempdir, 'out.ps'))
        # rsys2graph(rsys, os.path.join(tempdir, 'out.tex'))
        # https://github.com/kjellmf/dot2tex/issues/48
    finally:
        shutil.rmtree(tempdir)



# -*- coding: utf-8 -*-
"""
This module implements the denisty parameterisation of aqueous sulfuric acid
of Myhre et al. from 1998.
"""
from __future__ import (absolute_import, division, print_function)

import numpy as np
import warnings

from ..util import NoConvergence


_data = np.array([[999.8426, 0.03345402, -0.005691304, 0, 0],
                  [547.2659, -5.300445, 0.01187671, 0.0005990008, 0],
                  [5262.95, 37.20445, 0.1201909, -0.004148594, 1.197973E-5],
                  [-62139.58, -287.767, -0.4064638, 0.01119488, 3.607768E-5],
                  [409029.3, 1270.854, 0.326971, -0.01377435, -2.633585E-5],
                  [-1596989, -3062.836, 0.1366499, 0.006373031, 0],
                  [3857411, 4083.714, -0.1927785, 0, 0],
                  [-5808064, -2844.401, 0, 0, 0],
                  [5301976, 809.1053, 0, 0, 0],
                  [-2682616, 0, 0, 0, 0],
                  [576428.8, 0, 0, 0, 0]])


def sulfuric_acid_density(w, T=None, T0=None, units=None, warn=True):
    """
    Density of sulfuric acid (kg/m³) as function of temperature (K)
    and mass fraction acid (w).

    Parameters
    ----------
    w: float
        Acid mass fraction (0.1 <= w <= 0.9)
    T: float
        Temperature (in Kelvin) (273 <= T <= 323) (default: 298.15)
    T0: float
        Value of T for 0 degree Celsius (default: 273.15)
    units: object (optional)
        object with attributes: kelvin, meter, kilogram
    warn: bool (default: True)
        Emit UserWarning when outside T or w range.

    Returns
    -------
    Density of sulfuric acid (float of kg/m³ if T is float and units is None)

    Examples
    --------
    >>> print('%d' % sulfuric_acid_density(.5, 293))
    1396

    References
    ----------
    Cathrine E. L. Myhre , Claus J. Nielsen ,* and Ole W. Saastad
        "Density and Surface Tension of Aqueous H2SO4 at Low Temperature"
        J. Chem. Eng. Data, 1998, 43 (4), pp 617–622
        http://pubs.acs.org/doi/abs/10.1021/je980013g
        DOI: 10.1021/je980013g
    """
    if units is None:
        K = 1
        m = 1
        kg = 1
    else:
        K = units.Kelvin
        m = units.meter
        kg = units.kilogram
    if T is None:
        T = 298.15*K
    m3 = m**3
    if T0 is None:
        T0 = 273.15*K
    t = T - T0
    if warn:
        if np.any(t < 0*K) or np.any(t > 50*K):
            warnings.warn("Temperature is outside range (0-50 degC)")
        if np.any(w < 0.1) or np.any(w > .9):
            warnings.warn("Mass fraction is outside range (0.1-0.9)")
    t_arr = np.array([float(t/K)**j for j in range(5)]).reshape((1, 5))
    w_arr = np.array([w**i for i in range(11)]).reshape((11, 1))
    return np.sum((t_arr*w_arr)*_data)*kg/m3  # Equation (2) in reference


# generated at doi2bib.org:
bibtex = """
@article{Myhre1998,
  doi = {10.1021/je980013g},
  url = {http://dx.doi.org/10.1021/je980013g},
  year  = {1998},
  month = {jul},
  publisher = {American Chemical Society ({ACS})},
  volume = {43},
  number = {4},
  pages = {617--622},
  author = {Cathrine E. L. Myhre and Claus J. Nielsen and Ole W. Saastad},
  title = {Density and Surface Tension of Aqueous H2{SO}4 at Low Temperature},
  journal = {Journal of Chemical {\&} Engineering Data}
}
"""


def density_from_concentration(conc, T=None, molar_mass=None,
                               rho_cb=sulfuric_acid_density,
                               units=None, atol=None, maxiter=50, **kwargs):
    """ Calculates the density of a solution from its concentration

    Given a function which calculates the density of a solution from the mass
    fraction of the solute, this function calculates (iteratively) the density
    of said solution for a given concentration.

    Parameters
    ----------
    conc: float (optionally with units)
        concentration (mol / m³)
    T: float (optionally with units)
        passed to ``rho_cb``
    molar_mass: float (optionally with units)
        Molar mass of solute
    rho_cb: callback
        callback with signature f(w, T, units=None) -> rho
        (default: :func:`sulfuric_acid_density`)
    units: object (optional)
        object with attributes: meter, kilogram, mol
    atol: float (optionally with units)
        convergence criterion for fixed-point iteration
        (default: 1e-3 kg/m³)
    maxiter: int
        maximum number of iterations (when exceeded a NoConvergence excpetion
        is raised)
    \*\*kwargs:
        keyword arguments passed onto ``rho_cb``

    Returns
    -------
    Density of sulfuric acid (float of kg/m³ if T is float and units is None)

    Examples
    --------
    >>> print('%d' % density_from_concentration(400, 293))
    1021

    Raises
    ------
    chempy.util.NoConvergence:
        When maxiter is exceeded

    """
    if units is None:
        m = 1
        kg = 1
        mol = 1
    else:
        m = units.meter
        kg = units.kilogram
        mol = units.mol
    kg_per_m3 = kg * m**-3
    if atol is None:
        atol = 1e-3 * kg_per_m3
    if molar_mass is None:
        molar_mass = (1.00794*2 + 32.066 + 4*15.9994)*1e-3 * kg / mol

    rho = 1100 * kg_per_m3
    delta_rho = float('inf') * kg_per_m3

    iter_idx = 0
    while atol < abs(delta_rho):
        # fixed point iteration
        new_rho = rho_cb(conc*molar_mass/rho, T, units=units, **kwargs)
        delta_rho = new_rho - rho
        rho = new_rho
        iter_idx += 1
        if iter_idx > maxiter:
            raise NoConvergence("maxiter exceeded")
    return rho



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from ..pyutil import defaultnamedtuple


def test_defaultnamedtuple():
    Point2 = defaultnamedtuple('Point2', 'x y', [10])
    p = Point2(3)
    assert p.x == 3 and p.y == 10

    Point3 = defaultnamedtuple('Point2', 'x y z', [10, 20])
    p = Point3(3)
    assert p.x == 3 and p.y == 10 and p.z == 20

    p = Point3(3, z=30)
    assert p.x == 3 and p.y == 10 and p.z == 30

    p = Point3(3, 4, 5)
    assert p.x == 3 and p.y == 4 and p.z == 5



# -*- coding: utf-8 -*-

import numpy as np

from chempy.chemistry import Substance, Reaction, ReactionSystem
from chempy.units import (
    default_units, SI_base_registry, get_derived_unit, allclose
)
from ..ode import get_odesys
from ..integrated import dimerization_irrev


def test_get_odesys_1():
    k = .2
    a = Substance('A')
    b = Substance('B')
    r = Reaction({'A': 1}, {'B': 1}, param=k)
    rsys = ReactionSystem([r], [a, b])
    odesys = get_odesys(rsys, include_params=True)
    c0 = {
        'A': 1.0,
        'B': 3.0,
    }
    t = np.linspace(0.0, 10.0)
    xout, yout, info = odesys.integrate(t, rsys.as_per_substance_array(c0))
    yref = np.zeros((t.size, 2))
    yref[:, 0] = np.exp(-k*t)
    yref[:, 1] = 4 - np.exp(-k*t)
    assert np.allclose(yout, yref)


def test_get_odesys__with_units():
    a = Substance('A')
    b = Substance('B')
    molar = default_units.molar
    second = default_units.second
    r = Reaction({'A': 2}, {'B': 1}, param=1e-3/molar/second)
    rsys = ReactionSystem([r], [a, b])
    odesys = get_odesys(rsys, include_params=True,
                        unit_registry=SI_base_registry)
    c0 = {
        'A': 13*default_units.mol / default_units.metre**3,
        'B': .2 * default_units.molar
    }
    conc_unit = get_derived_unit(SI_base_registry, 'concentration')
    t = np.linspace(0, 10)*default_units.hour
    xout, yout, info = odesys.integrate(
        t, rsys.as_per_substance_array(c0, unit=conc_unit))
    Aref = dimerization_irrev(3600, 1e-6, 13.0)
    yref = np.zeros((xout.size, 2))
    yref[:, 0] = Aref
    yref[:, 1] = .2e-3 + 2*Aref
    print(yout)
    assert allclose(yout, yref*conc_unit)


def test_get_odesys_2():
    M = default_units.molar
    s = default_units.second
    substances = list(map(Substance, 'H2O H+ OH-'.split()))
    dissociation = Reaction({'H2O': 1}, {'H+': 1, 'OH-': 1}, 2.47e-5/s)
    recombination = Reaction({'H+': 1, 'OH-': 1}, {'H2O': 1}, 1.37e11/M/s)
    rsys = ReactionSystem([dissociation, recombination], substances)
    get_odesys(rsys, include_params=True,
               unit_registry=SI_base_registry, output_conc_unit=M)



from ..core import _ActivityProductBase, ionic_strength


def test_ionic_strength():
    assert abs(ionic_strength([0.1, 1.3, 2.1, 0.7],
                              [-1, 2, -3, 4]) - 17.7) < 1e-14


def test_ActivityProductBase():
    ap = _ActivityProductBase((1, -2, 3), 17, 42)
    assert ap.stoich == (1, -2, 3)
    assert ap.args == (17, 42)



# -*- coding: utf-8 -*-
"""
Module collecting classes and functions for dealing with (multiphase) chemical
equilibria.

.. Note::

  This module is provisional at the moment, i.e. the API is not stable and may
  break without a deprecation cycle.

"""
from __future__ import division, absolute_import

import math
import warnings
from collections import defaultdict
from functools import reduce
from operator import mul, add


import numpy as np

from .chemistry import ReactionSystem, equilibrium_quotient
from .util.pyutil import deprecated


def reducemap(args, reduce_op, map_op):
    return reduce(reduce_op, map(map_op, *args))


def vec_dot_vec(vec1, vec2):
    # return np.dot(vec1, vec2)
    # return np.add.reduce(np.multiply(vec1, vec2))
    return reducemap((vec1, vec2), add, mul)


def prodpow(bases, exponents):
    exponents = np.asarray(exponents)
    return np.multiply.reduce(bases**exponents, axis=-1)


def mat_dot_vec(iter_mat, iter_vec, iter_term=None):  # pure python (slow)
    if iter_term is None:
        return [vec_dot_vec(row, iter_vec) for row in iter_mat]
    else:
        # daxpy
        return [vec_dot_vec(row, iter_vec) + term for row, term
                in zip(iter_mat, iter_term)]


def equilibrium_residual(rc, c0, stoich, K, activity_product=None):
    """
    Parameters
    ---------
    rc: float
        Reaction coordinate
    c0: array_like of reals
        concentrations
    stoich: tuple
        per specie stoichiometry coefficient
    K: float
        equilibrium constant
    activity_product: callable
        callback for calculating the activity product taking
        concentration as single parameter.
    """
    if not hasattr(stoich, 'ndim') or stoich.ndim == 1:
        c = c0 + stoich*rc
    else:
        c = c0 + np.dot(stoich, rc)
    Q = equilibrium_quotient(c, stoich)
    if activity_product is not None:
        Q *= activity_product(c)
    return K - Q


def get_rc_interval(stoich, c0):
    """ get reaction coordinate interval """
    limits = c0/stoich
    if np.any(limits < 0):
        upper = -np.max(limits[np.argwhere(limits < 0)])
    else:
        upper = 0

    if np.any(limits > 0):
        lower = -np.min(limits[np.argwhere(limits > 0)])
    else:
        lower = 0

    if lower is 0 and upper is 0:
        raise ValueError("0-interval")
    else:
        return lower, upper


def _solve_equilibrium_coord(c0, stoich, K, activity_product=None):
    from scipy.optimize import brentq
    mask, = np.nonzero(stoich)
    stoich_m = stoich[mask]
    c0_m = c0[mask]
    lower, upper = get_rc_interval(stoich_m, c0_m)
    # span = upper - lower
    return brentq(
        equilibrium_residual,
        lower,  # + delta_frac*span,
        upper,  # - delta_frac*span,
        (c0_m, stoich_m, K, activity_product)
    )


def solve_equilibrium(c0, stoich, K, activity_product=None):
    """
    Solve equilibrium concentrations by using scipy.optimize.brentq

    Parameters
    ----------
    c0: array_like
        Initial guess of equilibrium concentrations
    stoich: tuple
        per specie stoichiometry coefficient (law of mass action)
    K: float
        equilibrium constant
    activity_product: callable
        see ``equilibrium_residual``
    delta_frac: float
        to avoid division by zero the span of searched values for
        the reactions coordinate (rc) is shrunk by 2*delta_frac*span(rc)
    """
    stoich = np.array(stoich)
    c0 = np.array(c0)
    rc = _solve_equilibrium_coord(c0, stoich, K, activity_product)
    return c0 + rc*stoich


def composition_balance(substances, concs, composition_number):
    if not hasattr(concs, 'ndim') or concs.ndim == 1:
        res = 0
    elif concs.ndim == 2:
        res = np.zeros(concs.shape[0])
        concs = concs.T
    else:
        raise NotImplementedError
    for s, c in zip(substances, concs):
        res += s.composition.get(composition_number, 0)*c
    return res


class _NumSys(object):

    small = 0  # precipitation limit
    pre_processor = None
    post_processor = None
    internal_x0_cb = None

    def __init__(self, eqsys, rref_equil=False, rref_preserv=False, ln=None,
                 exp=None, precipitates=()):
        if ln is None:
            ln = math.log
        if exp is None:
            exp = math.exp
        self.eqsys = eqsys
        self.rref_equil = rref_equil
        self.rref_preserv = rref_preserv
        self.ln = ln
        self.exp = exp
        self.precipitates = precipitates

    def _get_A_ks(self, eq_params):
        non_precip_rids = self.eqsys.non_precip_rids(self.precipitates)
        return self.eqsys.stoichs_constants(
            self.eqsys.eq_constants(non_precip_rids, eq_params, self.small),
            self.rref_equil, ln=self.ln, exp=self.exp,
            non_precip_rids=non_precip_rids)

    def _inits_and_eq_params(self, params):
        return params[:self.eqsys.ns], params[self.eqsys.ns:]


class NumSysLin(_NumSys):

    def internal_x0_cb(self, init_concs, params):
        # reduce risk of stationary starting point
        return (99*init_concs + self.eqsys.dissolved(init_concs))/100

    def f(self, yvec, params):
        from pyneqsys.symbolic import linear_exprs
        init_concs, eq_params = self._inits_and_eq_params(params)
        A, ks = self._get_A_ks(eq_params)
        # yvec == C
        f_equil = [q/k - 1 if k != 0 else q for q, k
                   in zip(prodpow(yvec, A), ks)]
        B, comp_nrs = self.eqsys.composition_balance_vectors()
        f_preserv = linear_exprs(B, yvec, mat_dot_vec(B, init_concs),
                                 rref=self.rref_preserv)
        return f_equil + f_preserv


class _NumSysLinNegPenalty(NumSysLin):

    def f(self, yvec, params):
        import sympy as sp
        f_penalty = [sp.Piecewise((yi**2, yi < 0), (0, True)) for yi in yvec]
        return super(_NumSysLinNegPenalty, self).f(yvec, params) + f_penalty


class NumSysLinRel(NumSysLin):

    def max_concs(self, params, min_=min, dtype=np.float64):
        init_concs = params[:self.eqsys.ns]
        return self.eqsys.upper_conc_bounds(init_concs, min_=min_, dtype=dtype)

    def pre_processor(self, x, params):
        return x/self.max_concs(params), params

    def post_processor(self, x, params):
        return x*self.max_concs(params), params

    def f(self, yvec, params):
        import sympy as sp
        return NumSysLin.f(self, [m*yi for m, yi in zip(
            self.max_concs(params, min_=lambda x: sp.Min(*x), dtype=object),
            yvec)], params)


class NumSysSquare(NumSysLin):

    small = 1e-35

    def pre_processor(self, x, params):
        return (np.sqrt(np.abs(x)), params)

    def post_processor(self, x, params):
        return x**2, params

    def internal_x0_cb(self, init_concs, params):
        return np.sqrt(np.abs(init_concs))

    def f(self, yvec, params):
        ysq = [yi*yi for yi in yvec]
        return NumSysLin.f(self, ysq, params)


class NumSysLinTanh(NumSysLin):

    def pre_processor(self, x, params):
        ymax = self.eqsys.upper_conc_bounds(params[:self.eqsys.ns])
        return np.arctanh((8*x/ymax - 4) / 5), params

    def post_processor(self, x, params):
        ymax = self.eqsys.upper_conc_bounds(params[:self.eqsys.ns])
        return ymax*(4 + 5*np.tanh(x))/8, params

    def internal_x0_cb(self, init_concs, params):
        return self.pre_processor(init_concs, init_concs)[0]

    def f(self, yvec, params):
        import sympy
        ymax = self.eqsys.upper_conc_bounds(
            params[:self.eqsys.ns],
            min_=lambda a, b: sympy.Piecewise((a, a < b), (b, True)))
        ytanh = [yimax*(4 + 5*sympy.tanh(yi))/8
                 for yimax, yi in zip(ymax, yvec)]
        return NumSysLin.f(self, ytanh, params)


class NumSysLog(_NumSys):

    small = math.exp(-80)  # anything less than `small` is insignificant

    def pre_processor(self, x, params):
        return (np.log(np.asarray(x) + NumSysLog.small),  # 10: damping
                params)  # zero conc. ~= small

    def post_processor(self, x, params):
        return np.exp(x), params

    def internal_x0_cb(self, init_concs, params):
        # return [1]*len(init_concs)
        return [0.1]*len(init_concs)

    def f(self, yvec, params):
        from pyneqsys.symbolic import linear_exprs
        init_concs, eq_params = self._inits_and_eq_params(params)
        A, ks = self._get_A_ks(eq_params)
        # yvec == ln(C)
        f_equil = mat_dot_vec(A, yvec, [-self.ln(k) for k in ks])
        B, comp_nrs = self.eqsys.composition_balance_vectors()
        f_preserv = linear_exprs(B, list(map(self.exp, yvec)),
                                 mat_dot_vec(B, init_concs),
                                 rref=self.rref_preserv)
        return f_equil + f_preserv


class EqSystem(ReactionSystem):

    def eq_constants(self, non_precip_rids=(), eq_params=None, small=0):
        if eq_params is None:
            eq_params = [eq.param for eq in self.rxns]
        return np.array([small if idx in non_precip_rids else
                         eq_params[idx] for idx, eq in enumerate(eq_params)])

    def upper_conc_bounds(self, init_concs, min_=min, dtype=np.float64):
        init_concs_arr = self.as_per_substance_array(init_concs, dtype=dtype)
        composition_conc = defaultdict(float)
        for conc, s_obj in zip(init_concs_arr, self.substances.values()):
            for comp_nr, coeff in s_obj.composition.items():
                if comp_nr == 0:  # charge may be created (if compensated)
                    continue
                composition_conc[comp_nr] += coeff*conc
        bounds = []
        for s_obj in self.substances.values():
            choose_from = []
            for comp_nr, coeff in s_obj.composition.items():
                if comp_nr == 0:
                    continue
                choose_from.append(composition_conc[comp_nr]/coeff)
            bounds.append(min_(choose_from))
        return bounds

    def equilibrium_quotients(self, concs):
        stoichs = self.stoichs()
        return [equilibrium_quotient(concs, stoichs[ri, :])
                for ri in range(self.nr)]

    def stoichs_constants(self, eq_params, rref=False, Matrix=None,
                          ln=None, exp=None, non_precip_rids=()):
        if rref:
            from pyneqsys.symbolic import linear_rref
            ln = ln or math.log
            rA, rb = linear_rref(self.stoichs(non_precip_rids),
                                 list(map(ln, eq_params)),
                                 Matrix)
            exp = exp or math.exp
            return rA.tolist(), list(map(exp, rb))
        else:
            return (self.stoichs(non_precip_rids),
                    eq_params)

    def composition_balance_vectors(self):
        composition_keys = set()
        for s in self.substances.values():
            for key in s.composition:
                composition_keys.add(key)
        vs = []
        sorted_composition_keys = sorted(composition_keys)
        for key in sorted_composition_keys:
            vs.append([s.composition.get(key, 0)
                       for s in self.substances.values()])
        return vs, sorted_composition_keys

    def composition_conservation(self, concs, init_concs):
        composition_vecs, comp_keys = self.composition_balance_vectors()
        A = np.array(composition_vecs)
        return (comp_keys,
                np.dot(A, self.as_per_substance_array(concs).T),
                np.dot(A, self.as_per_substance_array(init_concs).T))

    def other_phase_species_idxs(self, phase_idx=0):
        return [idx for idx, s in enumerate(
            self.substances.values()) if s.phase_idx != phase_idx]

    @property
    @deprecated(last_supported_version='0.3.1', will_be_missing_in='0.4.0',
                use_instead=other_phase_species_idxs)
    def precipitate_substance_idxs(self):
        return [idx for idx, s in enumerate(
            self.substances.values()) if s.precipitate]

    def phase_transfer_reaction_idxs(self, phase_idx=0):
        return [idx for idx, rxn in enumerate(self.rxns)
                if rxn.has_precipitates(self.substances)]

    @property
    @deprecated(last_supported_version='0.3.1', will_be_missing_in='0.4.0',
                use_instead=phase_transfer_reaction_idxs)
    def precipitate_rxn_idxs(self):
        return [idx for idx, rxn in enumerate(self.rxns)
                if rxn.has_precipitates(self.substances)]

    def dissolved(self, concs):
        """ Return dissolved concentrations """
        new_concs = concs.copy()
        for r in self.rxns:
            if r.has_precipitates(self.substances):
                net_stoich = np.asarray(r.net_stoich(self.substances))
                s_net, s_stoich, s_idx = r.precipitate_stoich(self.substances)
                new_concs -= new_concs[s_idx]/s_stoich * net_stoich
        return new_concs

    def _fw_cond_factory(self, ri):
        rxn = self.rxns[ri]

        def fw_cond(x, p):
            precip_stoich_coeff = rxn.precipitate_stoich(self.substances)[1]
            q = rxn.Q(self.substances, self.dissolved(x))
            k = rxn.K()
            QoverK = q/k
            if precip_stoich_coeff > 0:
                return QoverK < 1
            elif precip_stoich_coeff < 0:
                return QoverK > 1
            else:
                raise NotImplementedError
        return fw_cond

    def _bw_cond_factory(self, ri, small):
        rxn = self.rxns[ri]

        def bw_cond(x, p):
            precipitate_idx = rxn.precipitate_stoich(self.substances)[2]
            if x[precipitate_idx] < small:
                return False
            else:
                return True
            # precip_stoich_coeff = rxn.precipitate_stoich(self.substances)[1]
            # q = rxn.Q(self.substances, self.dissolved(x))
            # k = rxn.K()
            # QoverK = q/k
            # if precip_stoich_coeff > 0:
            #     return QoverK <= 1
            # elif prec_stoich_coeff < 0:
            #     return QoverK >= 1
            # else:
            #     raise NotImplementedError
        return bw_cond

    def _SymbolicSys_from_NumSys(self, NS, conds, rref_equil, rref_preserv):
        from pyneqsys.symbolic import SymbolicSys
        import sympy as sp
        ns = NS(self, ln=sp.log, exp=sp.exp, rref_equil=rref_equil,
                rref_preserv=rref_preserv, precipitates=conds)
        symb_kw = {}
        if ns.pre_processor is not None:
            symb_kw['pre_processors'] = [ns.pre_processor]
        if ns.post_processor is not None:
            symb_kw['post_processors'] = [ns.post_processor]
        if ns.internal_x0_cb is not None:
            symb_kw['internal_x0_cb'] = ns.internal_x0_cb
        return SymbolicSys.from_callback(
            ns.f, self.ns, nparams=self.ns + self.nr, **symb_kw)

    def get_neqsys_conditional_chained(self, init_concs, rref_equil=False,
                                       rref_preserv=False, NumSys=NumSysLin):
        from pyneqsys import ConditionalNeqSys, ChainedNeqSys

        def factory(conds):
            return ChainedNeqSys([self._SymbolicSys_from_NumSys(
                NS, conds, rref_equil, rref_preserv) for NS in NumSys])

        cond_cbs = [(self._fw_cond_factory(ri),
                     self._bw_cond_factory(ri, NumSys[0].small)) for
                    ri in self.phase_transfer_reaction_idxs()]
        return ConditionalNeqSys(cond_cbs, factory)

    def get_neqsys_chained_conditional(self, init_concs, rref_equil=False,
                                       rref_preserv=False,
                                       NumSys=NumSysLin):
        from pyneqsys import ConditionalNeqSys, ChainedNeqSys

        def mk_factory(NS):
            def factory(conds):
                return self._SymbolicSys_from_NumSys(NS, conds, rref_equil,
                                                     rref_preserv)
            return factory

        return ChainedNeqSys(
            [ConditionalNeqSys(
                [(self._fw_cond_factory(ri),
                  self._bw_cond_factory(ri, NS.small)) for
                 ri in self.phase_transfer_reaction_idxs()],
                mk_factory(NS)
            ) for NS in NumSys])

    def get_neqsys_static_conditions(self, init_concs, rref_equil=False,
                                     rref_preserv=False,
                                     NumSys=NumSysLin, precipitates=None):
        if precipitates is None:
            precipitates = (False,)*len(self.phase_transfer_reaction_idxs())
        from pyneqsys import ChainedNeqSys
        return ChainedNeqSys([self._SymbolicSys_from_NumSys(
            NS, precipitates, rref_equil, rref_preserv) for NS in NumSys])

    def get_neqsys(self, neqsys_type, init_concs, NumSys=NumSysLin, **kwargs):
        new_kw = {'rref_equil': False, 'rref_preserv': False}
        if neqsys_type == 'static_conditions':
            new_kw['precipitates'] = None
        for k in new_kw:
            if k in kwargs:
                new_kw[k] = kwargs.pop(k)

        try:
            NumSys[0]
        except TypeError:
            new_kw['NumSys'] = (NumSys,)
        else:
            new_kw['NumSys'] = NumSys

        return getattr(self, 'get_neqsys_' + neqsys_type)(init_concs, **new_kw)

    def non_precip_rids(self, precipitates):
        return [idx for idx, precip in zip(
            self.phase_transfer_reaction_idxs(), precipitates) if not precip]

    def _result_is_sane(self, init_concs, x):
        sc_upper_bounds = np.array(self.upper_conc_bounds(init_concs))
        neg_conc, too_much = np.any(x < 0), np.any(
            x > sc_upper_bounds*(1 + 1e-12))
        if neg_conc or too_much:
            if neg_conc:
                warnings.warn("Negative concentration")
            if too_much:
                warnings.warn("Too much of at least one component")
            return False
        return True

    def root(self, init_concs, x0=None, neqsys=None, NumSys=NumSysLog,
             neqsys_type='chained_conditional', **kwargs):
        init_concs = self.as_per_substance_array(init_concs)
        params = np.concatenate((init_concs, [float(elem) for elem
                                              in self.eq_constants()]))
        if neqsys is None:
            neqsys = self.get_neqsys(
                neqsys_type, init_concs, NumSys=NumSys,
                rref_equil=kwargs.pop('rref_equil', False),
                rref_preserv=kwargs.pop('rref_preserv', False),
                precipitates=kwargs.pop('precipitates', None))
        if x0 is None:
            x0 = init_concs
        x, sol = neqsys.solve(x0, params, **kwargs)
        if not sol['success']:
            warnings.warn("Root finding indicated as failed by solver.")
        sane = self._result_is_sane(init_concs, x)
        return x, sol, sane

    @staticmethod
    def _get_default_plot_ax(subplot_kwargs=None):
        import matplotlib.pyplot as plt
        if subplot_kwargs is None:
            subplot_kwargs = dict(xscale='log', yscale='log')
        return plt.subplot(1, 1, 1, **subplot_kwargs)

    def substance_labels(self, latex=False):
        if latex:
            result = ['$' + s.latex_name + '$'
                      for s in self.substances.values()]
            return result
        else:
            return [s.name for s in self.substances.values()]

    def roots(self, init_concs, varied_data, varied, x0=None,
              NumSys=NumSysLog, plot_kwargs=None,
              neqsys_type='chained_conditional', **kwargs):
        """
        Parameters
        ----------
        init_concs: array or dict
        varied_data: array
        varied_idx: int or str
        x0: array
        NumSys: _NumSys subclass
            See :class:`NumSysLin`, :class:`NumSysLog`, etc.
        plot_kwargs: dict
            See py:meth:`pyneqsys.NeqSys.solve`. Two additional keys
            are intercepted here:
                latex_names: bool (default: False)
                conc_unit_str: str (default: 'M')
        neqsys_type: str
            what method to use for NeqSys construction (get_neqsys_*)
        \*\*kwargs:
            kwargs passed on to py:meth:`pyneqsys.NeqSys.solve_series`
        """
        _plot = plot_kwargs is not None
        if _plot:
            latex_names = plot_kwargs.pop('latex_names', False)
            conc_unit_str = plot_kwargs.pop('conc_unit_str', 'M')
            if 'ax' not in plot_kwargs:
                plot_kwargs['ax'] = self._get_default_plot_ax()

        init_concs = self.as_per_substance_array(init_concs)
        neqsys = self.get_neqsys(
            neqsys_type, init_concs, NumSys=NumSys,
            rref_equil=kwargs.pop('rref_equil', False),
            rref_preserv=kwargs.pop('rref_preserv', False),
            precipitates=kwargs.pop('precipitates', None))
        if x0 is None:
            x0 = init_concs

        if _plot:
            cb = neqsys.solve_and_plot_series
            if 'plot_kwargs' not in kwargs:
                kwargs['plot_kwargs'] = {}
            if 'labels' not in kwargs['plot_kwargs']:
                kwargs['plot_kwargs']['labels'] = (
                    self.substance_labels(latex_names))
        else:
            cb = neqsys.solve_series

        params = np.concatenate((init_concs, self.eq_constants()))
        xvecs, info_dicts = cb(
            x0, params, varied_data, self.as_substance_index(varied),
            propagate=False, **kwargs)
        sanity = [self._result_is_sane(init_concs, x) for x in xvecs]

        if _plot:
            import matplotlib.pyplot as plt
            from pyneqsys.plotting import mpl_outside_legend
            mpl_outside_legend(plt.gca())
            varied_subst = self.substances[varied]
            xlbl = ('$[' + varied_subst.latex_name + ']_0$' if latex_names
                    else str(varied_subst))
            plt.gca().set_xlabel(xlbl + ' / ' + conc_unit_str)
            plt.gca().set_ylabel('Concentration / ' + conc_unit_str)

        return xvecs, info_dicts, sanity

    def plot_errors(self, concs, init_concs, varied_data, varied, axes=None,
                    compositions=True, Q=True, subplot_kwargs=None):
        if axes is None:
            import matplotlib.pyplot as plt
            if subplot_kwargs is None:
                subplot_kwargs = dict(xscale='log')
            fig, axes = plt.subplots(1, 2, figsize=(10, 4),
                                     subplot_kw=subplot_kwargs)
        varied_idx = self.as_substance_index(varied)
        ls, c = '- -- : -.'.split(), 'krgbcmy'
        all_inits = np.tile(self.as_per_substance_array(init_concs),
                            (len(varied_data), 1))
        all_inits[:, varied_idx] = varied_data
        if compositions:
            cmp_nrs, m1, m2 = self.composition_conservation(concs, all_inits)
            for cidx, (cmp_nr, a1, a2) in enumerate(zip(cmp_nrs, m1, m2)):
                axes[0].plot(concs[:, varied_idx],
                             a1-a2, label='Comp ' + str(cmp_nr),
                             ls=ls[cidx % len(ls)], c=c[cidx % len(c)])
                axes[1].plot(concs[:, varied_idx],
                             (a1-a2)/np.abs(a2), label='Comp ' + str(cmp_nr),
                             ls=ls[cidx % len(ls)], c=c[cidx % len(c)])

        if Q:
            # TODO: handle precipitate phases in plotting Q error
            qs = self.equilibrium_quotients(concs)
            ks = [rxn.param for rxn in self.rxns]
            for idx, (q, k) in enumerate(zip(qs, ks)):
                axes[0].plot(concs[:, varied_idx],
                             q-k, label='K R:' + str(idx),
                             ls=ls[(idx+cidx) % len(ls)],
                             c=c[(idx+cidx) % len(c)])
                axes[1].plot(concs[:, varied_idx],
                             (q-k)/k, label='K R:' + str(idx),
                             ls=ls[(idx+cidx) % len(ls)],
                             c=c[(idx+cidx) % len(c)])

        from pyneqsys.plotting import mpl_outside_legend
        mpl_outside_legend(axes[0])
        mpl_outside_legend(axes[1])
        axes[0].set_title("Absolute errors")
        axes[1].set_title("Relative errors")



"""
This module collects special cases of systems for which analytic solutions
are available.
"""
from __future__ import (absolute_import, division, print_function)

# Add documentation
# Rename esoteric parameter names


def dimerization_irrev(t, kf, initial_C, P0=1, t0=0):
    return 1/(1/initial_C + kf*(t-t0))


def pseudo_irrev(t, kf, P0, t0, excess_C, limiting_C, eps_l, exp=None):
    if exp is None:
        try:
            from numpy import exp
        except ImportError:
            from math import exp
    return P0*eps_l*limiting_C*(1 - exp(-excess_C*kf*(t-t0)))
pseudo_irrev.name = 'Pseudo first order irreversible'


def pseudo_rev(t, kf, P0, t0, excess_C, limiting_C, eps_l, beta, exp=None):
    if exp is None:
        try:
            from numpy import exp
        except ImportError:
            from math import exp
    kb = kf/beta
    return P0*eps_l*limiting_C*excess_C*kf/(excess_C*kf + kb)*(
        1 - exp(-(excess_C*kf+kb)*(t-t0)))
pseudo_rev.name = 'Pseudo first order reversible'


def binary_irrev(t, kf, P0, t0, excess_C, limiting_C, eps_l, exp=None):
    if exp is None:
        try:
            from numpy import exp
        except ImportError:
            from math import exp
    return P0*eps_l*excess_C*(1 - exp(-kf*(excess_C-limiting_C)*(t-t0)))/(
        excess_C/limiting_C - exp(-kf*(t-t0)*(excess_C-limiting_C)))
binary_irrev.name = 'Second order irreversible'


def binary_rev(t, kf, P0, t0, excess_C, limiting_C, eps_l, beta,
               exp=None, one=1):
    if exp is None:
        try:
            from numpy import exp
        except ImportError:
            from math import exp
    kb = kf/beta
    a = kf
    b = -excess_C*kf - limiting_C*kf - kb
    c = excess_C*limiting_C*kf
    P = (b**2 - 4*a*c)**(one/2)
    Q = P + b
    R = P - b
    return P0*eps_l*Q*(1 - exp(P*(t-t0)))/(2*a*(Q/R + exp(P*(t-t0))))
binary_rev.name = 'Second order reversible'



# -*- coding: utf-8 -*-
"""
ChemPy is a Python package which aims to be useful in solving chemistry
related problems.
"""

from __future__ import absolute_import, division, print_function

from ._url import __url__
from ._release import __version__
from .chemistry import Substance, Reaction, Equilibrium, ReactionSystem, Species
from .equilibria import EqSystem
from .henry import Henry
from .util.parsing import atomic_number



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function)

from ..arrhenius import arrhenius_equation


def test_arrhenius_equation():
    assert abs(arrhenius_equation(3, 831.4472, 100) - 3/2.7182818) < 1e-7



# -*- coding: utf-8 -*-
"""
Contains functions for the `Arrhenius equation
<https://en.wikipedia.org/wiki/Arrhenius_equation>`_
(:func:`arrhenius_equation`) and a convenience fitting routine
(:func:`fit_arrhenius_equation`).
"""
from __future__ import (absolute_import, division, print_function)


def _get_R(constants=None, units=None):
    if constants is None:
        R = 8.314472
        if units is not None:
            J = units.Joule
            K = units.Kelvin
            mol = units.mol
            R *= J/mol/K
    else:
        R = constants.molar_gas_constant
    return R


def arrhenius_equation(A, Ea, T, constants=None, units=None, exp=None):
    """
    Returns the rate coefficient according to the Arrhenius equation

    Parameters
    ----------
    A: float with unit
        frequency factor
    Ea: float with unit
        activation energy
    T: float with unit
        temperature
    constants: object (optional, default: None)
        if None:
            T assumed to be in Kelvin, Ea in J/(K mol)
        else:
            attributes accessed: molar_gas_constant
            Tip: pass quantities.constants
    units: object (optional, default: None)
        attributes accessed: Joule, Kelvin and mol
    exp: callback (optional)
        callback for calculating the exponential, default: numpy.exp, math.exp

    """
    if exp is None:
        try:
            from numpy import exp
        except ImportError:
            from math import exp
    R = _get_R(constants, units)
    try:
        RT = (R*T).rescale(Ea.dimensionality)
    except AttributeError:
        RT = R*T
    return A*exp(-Ea/RT)


def fit_arrhenius_equation(k, T, kerr=None, linearized=False):
    """ Curve fitting of the Arrhenius equation to data points

    Parameters
    ----------
    k: array_like
    T: float
    kerr: array_like (optional)
    linearized: bool

    """

    if len(k) != len(T):
        raise ValueError("k and T needs to be of equal length.")
    from math import exp
    import numpy as np
    rT = 1/T
    lnk = np.log(k)
    p = np.polyfit(rT, lnk, 1)
    R = _get_R(constants=None, units=None)
    Ea = -R*p[0]
    A = exp(p[1])
    if linearized:
        return A, Ea
    from scipy.optimize import curve_fit
    if kerr is None:
        weights = None
    else:
        weights = 1/kerr**2
    popt, pcov = curve_fit(arrhenius_equation, T, k, [A, Ea], weights)
    return popt, pcov



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import warnings


class Deprecation(object):
    """ factory of decorators for deprecating functions or classes

    Search the code base for more examples. Note that the wrapper gets
    :attr:`_deprecation` attribute with the :class:`Deprecation` instance.

    Parameters
    ----------
    last_supported_version : str
        Version string, e.g. ``'0.2.1'``.
    will_be_missing_in : str
        Version string, e.g. ``'0.3.0'``.
    use_instead : object or str
        Function or class to use instead or descriptive string.
    issue : str
        Issue identifier, e.g. 'gh-15' where 15 is the issue number on github.
        Note that 'gh-' prefix is a special case and stripped away.
    issues_url : str
        Url temlpate, e.g. ``'https://github.com/user/repo/issues/%s/'``.
    warning: DeprecationWarning
        Any subclass of DeprecationWarning, tip: you may invoke:
        ``warnings.simplefilter('once', MyWarning)`` at module init.

    Examples
    --------
    >>> import warnings
    >>> warnings.simplefilter("error", DeprecationWarning)
    >>> @Deprecation()
    ... def f():
    ...     return 1
    ...
    >>> f()  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    DeprecationWarning: f is deprecated.
    >>> @Deprecation(last_supported_version='0.4.0')
    ... def some_old_function(x):
    ...     return x*x - x
    ...
    >>> some_old_function._deprecation.last_supported_version
    '0.4.0'
    >>> @Deprecation(will_be_missing_in='1.0')
    ... class ClumsyClass(object):
    ...     pass
    ...
    >>> ClumsyClass._deprecation.will_be_missing_in
    '1.0'
    >>> warnings.resetwarnings()

    Notes
    -----
    :class:`DeprecationWarning` is ignored by default. Use custom warning
    and filter appropriately. Alternatively, run python with ``-W`` flag or set
    the appropriate environment variable:

    ::

        $ python -c 'import warnings as w; w.warn("X", DeprecationWarning)'
        $ python -Wd -c 'import warnings as w; w.warn("X", DeprecationWarning)'
        -c:1: DeprecationWarning: X
        $ export PYTHONWARNINGS=d
        $ python -c 'import warnings as w; w.warn("X", DeprecationWarning)'
        -c:1: DeprecationWarning: X


    """

    def __init__(self, last_supported_version=None, will_be_missing_in=None,
                 use_instead=None, issue=None, issues_url='%s',
                 warning=DeprecationWarning):
        self.last_supported_version = last_supported_version
        self.will_be_missing_in = will_be_missing_in
        self.use_instead = use_instead
        self.issue = issue
        self.issues_url = issues_url
        self.warning = warning
        self.warning_message = self._warning_message_template()

    def _warning_message_template(self):
        msg = '%(func_name)s is deprecated'
        if self.last_supported_version is not None:
            msg += ' since (not including) % s' % self.last_supported_version
        if self.will_be_missing_in is not None:
            msg += ', it will be missing in %s' % self.will_be_missing_in
        if self.issue is not None:
            if self.issue.startswith('gh-'):
                msg += ' (see %s)' % (self.issues_url % self.issue[3:])
            else:
                msg += ' (see %s)' % (self.issues_url % self.issue)
        if self.use_instead is not None:
            try:
                msg += '. Use %s instead' % self.use_instead.__name__
            except AttributeError:
                msg += '. Use %s instead' % self.use_instead
        return msg + '.'

    def __call__(self, wrapped):
        """ Decorates function to be deprecated """
        msg = self.warning_message % {'func_name': wrapped.__name__}
        wrapped_doc = wrapped.__doc__ or ''
        if hasattr(wrapped, '__mro__'):  # wrapped is a class
            class _Wrapper(wrapped):
                __doc__ = msg + '\n\n' + wrapped_doc

                def __init__(_self, *args, **kwargs):
                    warnings.warn(msg, self.warning, stacklevel=3)
                    wrapped.__init__(_self, *args, **kwargs)

        else:  # wrapped is a function
            def _Wrapper(*args, **kwargs):
                warnings.warn(msg, self.warning, stacklevel=3)
                return wrapped(*args, **kwargs)
            _Wrapper.__doc__ = msg + '\n\n' + wrapped_doc

        _Wrapper._deprecation = self
        _Wrapper.__name__ = wrapped.__name__
        _Wrapper.__module__ = wrapped.__module__
        return _Wrapper



from __future__ import division

from ..integrated import pseudo_irrev, pseudo_rev, binary_irrev, binary_rev

import sympy

one = sympy.S(1)

t, kf, P0, t0, excess_C, limiting_C, eps_l, beta = sympy.symbols(
    't k_f P0 t0 Y Z epsilon beta', negative=False)  # t0 => -t0

subsd = {t: one*2, kf: one*3, P0: one*5, t0: one*7, excess_C: one*11,
         limiting_C: one*13, eps_l: one*17, beta: one*23}
excl_params = {t0: one*0, P0: one, eps_l: one}


def test_pseudo_irrev():
    f = pseudo_irrev(t, kf, P0, -t0, excess_C, limiting_C, eps_l,
                     exp=sympy.exp).subs(excl_params)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    assert (num_dfdt - (
        excess_C*kf*(limiting_C - f)
    ).subs(subsd)).simplify() == 0


def test_pseudo_rev():
    f = pseudo_rev(t, kf, P0, -t0, excess_C, limiting_C, eps_l,
                   beta, exp=sympy.exp).subs(excl_params)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    assert (num_dfdt - (
        excess_C*kf*(limiting_C - f) - kf/beta*f
    ).subs(subsd)).simplify() == 0


def test_binary_irrev():
    f = binary_irrev(t, kf, P0, -t0, excess_C, limiting_C, eps_l,
                     exp=sympy.exp).subs(excl_params)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    assert (num_dfdt - (
        kf*(limiting_C - f)*(excess_C - f)
    ).subs(subsd)).simplify() == 0


def test_binary_rev():
    f = binary_rev(t, kf, P0, -t0, excess_C, limiting_C, eps_l,
                   beta, exp=sympy.exp).subs(excl_params)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    ans = kf*(limiting_C - f)*(excess_C - f) - kf/beta*f
    # symbolic susbsitution fails:
    assert abs(float(num_dfdt) - float(ans.subs(subsd))) < 2e-14



# -*- coding: utf-8 -*-
""" The units module provides the following attributes:

- ``chempy.units.default_units``
- ``chempy.units.default_constants``
- ``chempy.units.SI_base_registry``

Currently `quantities <https://pypi.python.org/pypi/quantities>`_ is used as
the underlying package to handle units. If it is possible you should try to
only use the `chempy.units` module (in case ``ChemPy`` changes this backend).

"""
from __future__ import (absolute_import, division, print_function)

# Currently we use quantities for units. This may change, therefore use this
# file for all units. A requirement is first-class numpy support.

import numpy as np
try:
    import quantities as pq
except ImportError:
    UncertainQuantity = None
    default_constants = None
    default_units = None
    SI_base_registry = None
else:
    UncertainQuantity = pq.UncertainQuantity
    # Let us extend the underlying pq namespace with some common units in
    # chemistry
    default_constants = pq.constants

    class _NameSpace:
        def __init__(self, default):
            self._NameSpace_default = default
            self._NameSpace_attr_store = {}

        def __getattr__(self, attr):
            if attr.startswith('_NameSpace_'):
                return self.__dict__[attr]
            else:
                try:
                    return self._NameSpace_attr_store[attr]
                except KeyError:
                    return getattr(self._NameSpace_default, attr)

        def __setattr__(self, attr, val):
            if attr.startswith('_NameSpace_'):
                self.__dict__[attr] = val
            else:
                self._NameSpace_attr_store[attr] = val

        def as_dict(self):
            result = {k: v for k, v in self._NameSpace_default.__dict__.items()
                      if not k.startswith('_')}
            result.update(self._NameSpace_attr_store)
            return result

    default_units = _NameSpace(pq)
    default_units.decimetre = pq.UnitQuantity(
        'decimetre',  default_units.m / 10.0, u_symbol='dm')
    if not hasattr(default_units, 'molar'):
        default_units.molar = pq.UnitQuantity(
            'M',  default_units.mole / default_units.decimetre ** 3,
            u_symbol='M')
    default_units.per100eV = pq.UnitQuantity(
        'per_100_eV',
        1/(100*default_units.eV*default_constants.Avogadro_constant),
        u_symbol='(100eV)**-1')
    default_units.micromole = pq.UnitQuantity(
        'micromole',  pq.mole/1e6,  u_symbol=u'μmol')
    default_units.kilojoule = pq.UnitQuantity(
        'kilojoule',  1e3*pq.joule,  u_symbol='kJ')
    default_units.perMolar_perSecond = 1/default_units.molar/pq.s
    default_units.per100eV = pq.UnitQuantity(
        'per_100_eV', 1/(100*pq.eV*pq.constants.Avogadro_constant),
        u_symbol='(100eV)**-1')
    default_units.umol = pq.UnitQuantity('micromole',  pq.mole/1e6,
                                         u_symbol=u'μmol')
    default_units.umol_per_J = default_units.umol / pq.joule

    # unit registry data and logic:

    SI_base_registry = {
        'length': default_units.metre,
        'mass': default_units.kilogram,
        'time': default_units.second,
        'current': default_units.ampere,
        'temperature': default_units.kelvin,
        'luminous_intensity': default_units.candela,
        'amount': default_units.mole
    }


def get_derived_unit(registry, key):
    """ Get the unit of a physcial quantity in a provided unit system.

    Parameters
    ----------
    registry: dict (str: unit)
        mapping 'length', 'mass', 'time', 'current', 'temperature',
        'luminous_intensity', 'amount'. If registry is ``None`` the
        function returns 1.0 unconditionally.
    key: str
        one of the registry keys or one of: 'diffusion', 'electrical_mobility',
        'permittivity', 'charge', 'energy', 'concentration', 'density',
        'radiolytic_yield'

    Examples
    --------
    >>> m, s = default_units.meter, default_units.second
    >>> get_derived_unit(SI_base_registry, 'diffusion') == m**2/s
    True

    """
    if registry is None:
        return 1.0
    derived = {
        'diffusion': registry['length']**2/registry['time'],
        'electrical_mobility': (registry['current']*registry['time']**2 /
                                registry['mass']),
        'permittivity': (registry['current']**2*registry['time']**4 /
                         (registry['length']**3*registry['mass'])),
        'charge': registry['current']*registry['time'],
        'energy': registry['mass']*registry['length']**2/registry['time']**2,
        'concentration': registry['amount']/registry['length']**3,
        'density': registry['mass']/registry['length']**3,
    }
    derived['radiolytic_yield'] = registry['amount']/derived['energy']
    try:
        return derived[key]
    except KeyError:
        return registry[key]


def unit_registry_to_human_readable(unit_registry):
    """ Serialization of a unit registry. """
    if unit_registry is None:
        return None
    new_registry = {}
    for k in SI_base_registry:
        if unit_registry[k] is 1:
            new_registry[k] = 1, 1
        else:
            dim_list = list(unit_registry[k].dimensionality)
            if len(dim_list) != 1:
                raise TypeError("Compound units not allowed: {}".format(
                    dim_list))
            u_symbol = dim_list[0].u_symbol
            # u_symbol = unit_registry[k].u_symbol
            new_registry[k] = float(unit_registry[k]), u_symbol
    return new_registry


def unit_registry_from_human_readable(unit_registry):
    """ Deserialization of unit_registry. """
    if unit_registry is None:
        return None
    new_registry = {}
    for k in SI_base_registry:
        factor, u_symbol = unit_registry[k]
        if u_symbol == 1:
            unit_quants = [1]
        else:
            unit_quants = list(pq.Quantity(0, u_symbol).dimensionality.keys())

        if len(unit_quants) != 1:
            raise TypeError("Unkown UnitQuantity: {}".format(unit_registry[k]))
        else:
            new_registry[k] = factor*unit_quants[0]
    return new_registry


# Abstraction of underlying package providing units and dimensional analysis:

def is_unitless(expr):
    """ Returns ``True`` if ``expr`` is unitless, otherwise ``False``

    Examples
    --------
    >>> is_unitless(42)
    True
    >>> is_unitless(42*default_units.kilogram)
    False

    """
    if hasattr(expr, 'dimensionality'):
        return expr.dimensionality == pq.dimensionless.dimensionality
    return True


def unit_of(expr):
    """ Returns the unit of a quantity

    Examples
    --------
    >>> unit_of(42*pq.second) == unit_of(12*pq.second)
    True
    >>> unit_of(42)
    1

    """
    try:
        return expr.units
    except AttributeError:
        return 1


def to_unitless(value, new_unit=None):
    """ Nondimensionalization of a quantity.

    Parameters
    ----------
    value: quantity
    new_unit: unit

    Examples
    --------
    >>> '%.1g' % to_unitless(1*default_units.metre, default_units.nm)
    '1e+09'

    """
    if new_unit is None:
        new_unit = pq.dimensionless
    if isinstance(value, (list, tuple)):
        return np.array([to_unitless(elem, new_unit) for elem in value])
    if isinstance(value, dict):
        new_value = value.copy()
        for k in new_value:
            new_value[k] = to_unitless(new_value[k], new_unit)
        return new_value
    try:
        result = (value*pq.dimensionless/new_unit).rescale(pq.dimensionless)
        if result.ndim == 0:
            return float(result)
        else:
            return np.asarray(result)
    except TypeError:
        return np.array([to_unitless(elem, new_unit) for elem in value])


# NumPy like functions for compatibility:

def allclose(a, b, rtol=1e-8, atol=None):
    """ Analogous to ``numpy.allclose``. """
    try:
        d = abs(a - b)
    except TypeError:
        if len(a) == len(b):
            return all(allclose(_a, _b, rtol, atol) for _a, _b in zip(a, b))
        raise
    lim = abs(a)*rtol
    if atol is not None:
        lim += atol
    try:
        n = len(d)
    except TypeError:
        n = 1

    if n == 1:
        return d < lim
    else:
        return np.all(_d < _lim for _d, _lim in zip(d, lim))


def linspace(start, stop, num=50):
    """ Analogous to ``numpy.linspace``. """
    # work around for quantities v0.10.1 and NumPy
    unit = unit_of(start)
    start_ = to_unitless(start, unit)
    stop_ = to_unitless(stop, unit)
    return np.linspace(start_, stop_, num)*unit



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import collections

import pytest
import numpy as np
from scipy.optimize import fsolve

from ..chemistry import (
    Substance, Reaction, Equilibrium, Species, Solute,
)

from ..equilibria import (
    equilibrium_quotient, equilibrium_residual, get_rc_interval,
    solve_equilibrium, prodpow, EqSystem, NumSysLin, NumSysLog
)

from .ammonical_cupric_solution import get_ammonical_cupric_eqsys


def test_equilibrium_quotient():
    assert abs(equilibrium_quotient([2.3, 3.7, 5.1], (-1, -1, 1)) -
               5.1/2.3/3.7) < 1e-14


def test_equilibrium_residual():
    c0 = np.array((13., 11, 17))
    stoich = np.array((-2, 3, -4))
    K = 3.14
    assert abs(equilibrium_residual(0.1, c0, stoich, K) -
               (K - (13-0.2)**-2*(11 + 0.3)**3*(17 - 0.4)**-4)) < 1e-14


def test_get_rc_interval():
    c = np.array((13., 11, 17))
    stoich = np.array((-2, 3, -4))
    limits = get_rc_interval(stoich, c)
    lower = -11/3.
    upper = 17./4
    assert abs(limits[0] - lower) < 1e-14
    assert abs(limits[1] - upper) < 1e-14


def test_solve_equilibrium_1():
    c = np.array((13., 11, 17))
    stoich = np.array((-2, 3, -4))
    K = 3.14

    def f(x):
        return (13 - 2*x)**-2 * (
            11 + 3*x)**3 * (17 - 4*x)**-4 - K
    assert np.allclose(solve_equilibrium(c, stoich, K),
                       c + stoich*fsolve(f, 3.48))


def test_solve_equilibrium_2():
    c = np.array([1.7e-03, 3.0e+06, 3.0e+06, 9.7e+07, 5.55e+09])
    stoich = (1, 1, 0, 0, -1)
    K = 55*1e-6

    def f(x):
        return prodpow(c+x*stoich, stoich) - K
    solution = solve_equilibrium(c, stoich, K)
    assert np.allclose(solution, c + stoich*fsolve(f, 0.1))


def test_EqSystem():
    a, b = sbstncs = Substance('a'), Substance('b')
    rxns = [Reaction({'a': 1}, {'b': 1})]
    es = EqSystem(rxns, [(s.name, s) for s in sbstncs])
    assert es.net_stoichs().tolist() == [[-1, 1]]


def _get_es1():
    a, b = sbstncs = Solute('a'), Solute('b')
    rxns = [Equilibrium({'a': 1}, {'b': 1}, 3)]
    return EqSystem(rxns, sbstncs)


def _get_es_water(EqSys=EqSystem):
    H2O = Substance('H2O', charge=0, composition={1: 2, 8: 1})
    OHm = Substance('OH-', charge=-1, composition={1: 1, 8: 1})
    Hp = Substance('H+', charge=1, composition={1: 1})
    Kw = 1e-14/55.5
    w_auto_p = Equilibrium({'H2O': 1}, {'Hp': 1, 'OHm': 1}, Kw)
    return EqSys([w_auto_p], [H2O, OHm, Hp])


def test_EqSystem_1():
    es = _get_es1()
    assert es.stoichs().tolist() == [[-1, 1]]
    assert es.eq_constants() == [3]


def test_Equilibria_arithmetics():
    es1 = _get_es1()
    e, = es1.rxns
    e2 = 2*e
    sum2 = e + e
    assert sum2 == e2


def test_Equilibria_root():
    eqsys, c0 = get_ammonical_cupric_eqsys()
    x, sol, sane = eqsys.root(c0, NumSys=(NumSysLog,))
    assert sol['success'] and sane


def _solutes(Cls):
    return (
        Cls('H2O', 0, composition={1: 2, 8: 1}),
        Cls('H+', 1, composition={1: 1}),
        Cls('OH-', -1, composition={1: 1, 8: 1}),
        Cls('NH4+', 1, composition={1: 4, 7: 1}),
        Cls('NH3', 0, composition={1: 3, 7: 1})
    )


@pytest.mark.parametrize('Cls', [Solute, Species])
def test_Equilibria_root_simple(Cls):
    solutes = water, hydronium, hydroxide, ammonium, ammonia = _solutes(Cls)

    water_auto_protolysis = Equilibrium(
        {water.name: 1}, {hydronium.name: 1, hydroxide.name: 1}, 1e-14/55.5)
    ammonia_protolysis = Equilibrium(
        {ammonium.name: 1}, {hydronium.name: 1, ammonia.name: 1},
        10**-9.26/55.5
    )
    eqsys = EqSystem([water_auto_protolysis, ammonia_protolysis], solutes)

    init_concs = collections.defaultdict(float, {
        water.name: 55.5, ammonia.name: 1e-3})
    x, sol1, sane1 = eqsys.root(init_concs)
    x, sol2, sane2 = eqsys.root(init_concs, x, NumSys=(NumSysLog,))
    assert sane2
    ref = eqsys.as_per_substance_array({
        water.name: 55.5,
        ammonia.name: 1e-3 - 6.2e-4,
        ammonium.name: 6.2e-4,
        hydronium.name: 1.6e-11,
        hydroxide.name: 6.2e-4
    })
    assert np.allclose(x, ref, rtol=0.02, atol=1e-16)


def _get_NaCl(Cls, **precipitate_kwargs):
    Na_p, Cl_m, NaCl = sbstncs = (
        Cls('Na+', 1, composition={11: 1}),
        Cls('Cl-', -1, composition={17: 1}),
        Cls('NaCl', composition={11: 1, 17: 1}, **precipitate_kwargs)
    )
    sp = Equilibrium({'NaCl': 1}, {'Na+': 1, 'Cl-': 1}, 4.0)
    eqsys = EqSystem([sp], sbstncs)
    cases = [
        [(0, 0, .1), (.1, .1, 0)],
        [(.5, .5, .4), (.9, .9, 0)],
        [(1, 1, 1), (2, 2, 0)],
        [(0, 0, 2), (2, 2, 0)],
        [(3, 3, 3), (2, 2, 4)],
        [(3, 3, 0), (2, 2, 1)],
        [(0, 0, 3), (2, 2, 1)],
    ]
    return eqsys, [s.name for s in sbstncs], cases


@pytest.mark.parametrize('kwargs', [
    dict(Cls=Solute, precipitate=True),
    dict(Cls=Species, phase_idx=1)
])
def test_EqSystem_dissolved(kwargs):
    eqsys, names, _ = _get_NaCl(**kwargs)
    inp = eqsys.as_per_substance_array({'Na+': 1, 'Cl-': 2, 'NaCl': 4})
    result = eqsys.dissolved(inp)
    ref = eqsys.as_per_substance_array({'Na+': 5, 'Cl-': 6, 'NaCl': 0})
    assert np.allclose(result, ref)


@pytest.mark.parametrize('NumSys', [(NumSysLin,), (NumSysLog,),
                                    (NumSysLog, NumSysLin)])
def test_precipitate(NumSys):
    eqsys, species, cases = _get_NaCl(Solute, precipitate=True)

    for init, final in cases:
        x, sol, sane = eqsys.root(dict(zip(species, init)),
                                  NumSys=NumSys, rref_preserv=True)
        assert sol['success'] and sane
        assert x is not None
        assert np.allclose(x, np.asarray(final))



# -*- coding: utf-8 -*-
"""
This modules collects expressions for the Debye-Hückel constants.

"""
from __future__ import (absolute_import, division, print_function)

from .core import _ActivityProductBase, ionic_strength


def A(eps_r, T, rho, b0=1, constants=None, units=None, one=1):
    """
    Debye Huckel constant A

    Parameters
    ----------
    eps_r: float
        relative permittivity
    T: float with unit
        Temperature (default: assume Kelvin)
    rho: float with unit
        density (default: assume kg/m**3)
    b0: float with unit
        reference molality (default: assume mol/kg)
    units: object (optional, default: None)
        attributes accessed: meter, Kelvin and mol
    constants: object (optional, default: None)
        if None:
            T assumed to be in Kelvin and b0 = 1 mol/kg
        else:
            see source code for what attributes are used.
            Tip: pass quantities.constants

    Notes
    -----
    Remember to divide by ln(10) if you want to use the constant
    with log10 based expression.

    References
    ----------
    Atkins, De Paula, Physical Chemistry, 8th edition

    """

    if constants is None:
        combined = 132871.85866393594
        if units is not None:
            m = units.meter
            K = units.Kelvin
            mol = units.mol
            combined *= (m*K)**(3*one/2) / mol**(one/2)
        return combined*(rho * b0 * T**-3 * eps_r**-3)**0.5
    F = constants.Faraday_constant
    NA = constants.Avogadro_constant
    eps0 = constants.vacuum_permittivity
    kB = constants.Boltzmann_constant
    pi = constants.pi
    A = F**3/(4*pi*NA)*(rho*b0/(2*(eps0*eps_r*kB*NA*T)**3))**(one/2)
    return A


def B(eps_r, T, rho, b0=1, constants=None, units=None, one=1):
    """
    Extended Debye-Huckel parameter B

    Parameters
    ----------
    eps_r: float
        relative permittivity
    T: float with unit
        temperature
    rho: float with unit
        density
    b0: float with unit
        reference molality
    units: object (optional, default: None)
        attributes accessed: meter, Kelvin and mol
    constants: object (optional, default: None)
        if None:
            T assumed to be in Kelvin, rho in kg/m**3 and b0 = 1 mol/kg
        else:
            attributes accessed: molar_gas_constant, Faraday_constant
            Tip: pass quantities.constants

    Returns
    -------
    Debye Huckel B constant (default in m**-1)

    """

    if constants is None:
        combined = 15903203868.740343
        if units is not None:
            m = units.meter
            K = units.Kelvin
            mol = units.mol
            combined *= (m*K/mol)**(one/2)
        return combined*(rho*b0/(T*eps_r))**0.5
    F = constants.Faraday_constant
    eps0 = constants.vacuum_permittivity
    R = constants.molar_gas_constant
    B = F*(2*rho*b0/(eps_r*eps0*R*T))**(one/2)
    return B


def limiting_log_gamma(I, z, A, I0=1, one=1):
    """ Debye-Hyckel limiting formula """
    # `one` allows passing of e.g. one=sympy.S(1)
    return -A*z**2*(I/I0)**(one/2)


def extended_log_gamma(I, z, a, A, B, C=0, I0=1, one=1):
    """ Debye-Huckel extended formula """
    # `one` allows passing of e.g. one=sympy.S(1)
    I_I0 = I/I0
    sqrt_I_I0 = (I_I0)**(one/2)
    return -A*z**2 * sqrt_I_I0/(1 + B*a*sqrt_I_I0) + C*I_I0


def davies_log_gamma(I, z, A, C=-0.3, I0=1, one=1):
    """ Davies formula """
    I_I0 = I/I0
    sqrt_I_I0 = (I_I0)**(one/2)
    return -A * z**2 * (sqrt_I_I0/(1 + sqrt_I_I0) + C*I_I0)


def limiting_activity_product(I, stoich, z, T, eps_r, rho, exp=None):
    if exp is None:
        try:
            from numpy import exp
        except ImportError:
            from math import exp
    Aval = A(eps_r, T, rho)
    tot = 0
    for idx, nr in enumerate(stoich):
        tot += nr*limiting_log_gamma(I, z[idx], Aval)
    return exp(tot)


def extended_activity_product(I, stoich, z, a, T, eps_r, rho, C=0, exp=None):
    if exp is None:
        try:
            from numpy import exp
        except ImportError:
            from math import exp
    Aval = A(eps_r, T, rho)
    Bval = B(eps_r, T, rho)
    tot = 0
    for idx, nr in enumerate(stoich):
        tot += nr*extended_log_gamma(I, z[idx], a[idx], Aval, Bval, C)
    return exp(tot)


def davies_activity_product(I, stoich, z, a, T, eps_r, rho, C=-0.3, exp=None):
    if exp is None:
        try:
            from numpy import exp
        except ImportError:
            from math import exp
    Aval = A(eps_r, T, rho)
    tot = 0
    for idx, nr in enumerate(stoich):
        tot += nr*davies_log_gamma(I, z[idx], Aval, C)
    return exp(tot)


class LimitingDebyeHuckelActivityProduct(_ActivityProductBase):

    def __call__(self, c):
        z = self.args[0]
        I = ionic_strength(c, z)
        return limiting_activity_product(I, self.stoich, *self.args)


class ExtendedDebyeHuckelActivityProduct(_ActivityProductBase):

    def __call__(self, c):
        z = self.args[0]
        I = ionic_strength(c, z)
        return extended_activity_product(I, self.stoich, *self.args)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)


_radii_nm = {  # in nanometer
    'Fe/3+': 0.9,
    'SCN/-': 0.35,
    'Fe/2+': 0.6,
}


def get_radii(key, units=None):
    """ Get Debye-Hückel radii for various ions

    For aqueous systems

    Parameters
    ----------
    key: str
        e.g. 'Fe/3+', 'SCN/-'
    units: object (optional)
        :attr:`nm` is accessed.

    Returns
    -------
    radius in nanometers

    """
    if units is None:
        return _radii_nm[key]
    else:
        return _radii_nm[key]*units.nm



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import pytest

from ..parsing import (
    to_composition, relative_atomic_masses, mass_from_composition,
    to_latex, to_reaction, atomic_number
)


def test_atomic_number():
    assert atomic_number('U') == 92
    assert atomic_number('carbon') == 6
    assert atomic_number('ununpentium') == 115
    with pytest.raises(ValueError):
        atomic_number('unobtainium')


def test_to_composition():
    assert to_composition('H2O') == {1: 2, 8: 1}
    assert to_composition('Fe/3+') == {0: 3, 26: 1}
    assert to_composition('Fe+3') == {0: 3, 26: 1}
    assert to_composition('Na/+') == {0: 1, 11: 1}
    assert to_composition('Na+1') == {0: 1, 11: 1}
    assert to_composition('Na+') == {0: 1, 11: 1}
    assert to_composition('Cl/-') == {0: -1, 17: 1}
    assert to_composition('Cl-') == {0: -1, 17: 1}
    assert to_composition('NaCl') == {11: 1, 17: 1}
    assert to_composition('NaCl(s)') == {11: 1, 17: 1}
    assert to_composition('Fe(SCN)2/+') == {0: 1, 6: 2, 7: 2, 16: 2, 26: 1}
    assert to_composition('Fe(SCN)2+') == {0: 1, 6: 2, 7: 2, 16: 2, 26: 1}
    assert to_composition('Fe(SCN)2+1') == {0: 1, 6: 2, 7: 2, 16: 2, 26: 1}
    assert to_composition('((H2O)2OH)12') == {1: 60, 8: 36}

    # Special case: solvated electron:
    assert to_composition('e-') == {0: -1}
    assert to_composition('e-1') == {0: -1}
    assert to_composition('e-(aq)') == {0: -1}

    # prefixes and suffixes
    assert to_composition('.NO2(g)') == {7: 1, 8: 2}
    assert to_composition('.NH2') == {1: 2, 7: 1}
    assert to_composition('ONOOH') == {1: 1, 7: 1, 8: 3}
    assert to_composition('.ONOO') == {7: 1, 8: 3}
    assert to_composition('.NO3/2-') == {0: -2, 7: 1, 8: 3}
    assert to_composition('.NO3-2') == {0: -2, 7: 1, 8: 3}

    with pytest.raises(ValueError):
        to_composition('F-F')

    assert to_composition('alpha-FeOOH(s)') == {1: 1, 8: 2, 26: 1}
    assert to_composition('epsilon-Zn(OH)2(s)') == {1: 2, 8: 2, 30: 1}


def test_relative_atomic_masses():
    assert relative_atomic_masses[0] == 1.008


def test_mass_from_composition():
    mass = mass_from_composition(to_composition('NaF'))
    assert abs(41.988172443 - mass) < 1e-7

    Fminus = mass_from_composition(to_composition('F/-'))
    assert abs(Fminus - 18.998403163 - 5.489e-4) < 1e-7


def test_to_latex():
    assert to_latex('H2O') == 'H_{2}O'
    assert to_latex('C6H6/+') == 'C_{6}H_{6}^{+}'
    assert to_latex('Fe(CN)6/3-') == 'Fe(CN)_{6}^{3-}'
    assert to_latex('Fe(CN)6-3') == 'Fe(CN)_{6}^{3-}'
    assert to_latex('C18H38/2+') == 'C_{18}H_{38}^{2+}'
    assert to_latex('C18H38/+2') == 'C_{18}H_{38}^{2+}'
    assert to_latex('C18H38+2') == 'C_{18}H_{38}^{2+}'
    assert to_latex('((H2O)2OH)12') == '((H_{2}O)_{2}OH)_{12}'
    assert to_latex('NaCl') == 'NaCl'
    assert to_latex('NaCl(s)') == 'NaCl(s)'
    assert to_latex('e-(aq)') == 'e^{-}(aq)'
    assert to_latex('.NO2(g)') == r'^\bullet NO_{2}(g)'
    assert to_latex('.NH2') == r'^\bullet NH_{2}'
    assert to_latex('ONOOH') == 'ONOOH'
    assert to_latex('.ONOO') == r'^\bullet ONOO'
    assert to_latex('.NO3/2-') == r'^\bullet NO_{3}^{2-}'
    assert to_latex('.NO3-2') == r'^\bullet NO_{3}^{2-}'
    assert to_latex('alpha-FeOOH(s)') == r'\alpha-FeOOH(s)'
    assert to_latex('epsilon-Zn(OH)2(s)') == r'\varepsilon-Zn(OH)_{2}(s)'


def test_to_reaction():
    from chempy.chemistry import Reaction, Equilibrium
    rxn = to_reaction(
        "H+ + OH- -> H2O; 1.4e11; ref={'doi': '10.1039/FT9908601539'}",
        'H+ OH- H2O'.split(), '->', Reaction)
    assert rxn.__class__ == Reaction

    assert rxn.reac['H+'] == 1
    assert rxn.reac['OH-'] == 1
    assert rxn.prod['H2O'] == 1
    assert rxn.param == 1.4e11
    assert rxn.ref['doi'].startswith('10.')

    eq = to_reaction("H+ + OH- = H2O; 1e-14; ref='rt, [H2O] == 1 M'",
                     'H+ OH- H2O'.split(), '=', Equilibrium)
    assert eq.__class__ == Equilibrium

    assert eq.reac['H+'] == 1
    assert eq.reac['OH-'] == 1
    assert eq.prod['H2O'] == 1
    assert eq.ref.startswith('rt')

    for s in ['2 e-(aq) + (2 H2O) -> H2 + 2 OH- ; 1e6 ; ',
              '2 * e-(aq) + (2 H2O) -> 1 * H2 + 2 * OH- ; 1e6 ; ']:
        rxn2 = to_reaction(s, 'e-(aq) H2 OH- H2O'.split(), '->', Reaction)
        assert rxn2.__class__ == Reaction
        assert rxn2.reac['e-(aq)'] == 2
        assert rxn2.inact_reac['H2O'] == 2
        assert rxn2.prod['H2'] == 1
        assert rxn2.prod['OH-'] == 2
        assert rxn2.param == 1e6



# -*- coding: utf-8 -*-
""" Functions for chemical formulae and reactions """

from __future__ import (absolute_import, division, print_function)

from collections import defaultdict

import re
import warnings

from pyparsing import (
    Forward, Group, OneOrMore, Optional, ParseResults,
    Regex, Suppress, Word, nums
)

from .pyutil import ChemPyDeprecationWarning


def _get_formula_parser():
    """ Create a forward pyparsing parser for chemical formulae

    BNF for simple chemical formula (no nesting)

        integer :: '0'..'9'+
        element :: 'A'..'Z' 'a'..'z'*
        term :: element [integer]
        formula :: term+


    BNF for nested chemical formula

        integer :: '0'..'9'+
        element :: 'A'..'Z' 'a'..'z'*
        term :: (element | '(' formula ')') [integer]
        formula :: term+

    Notes
    -----
    The code in this function is from an answer on StackOverflow:
        http://stackoverflow.com/a/18555142/790973
        written by:
            Paul McGuire, http://stackoverflow.com/users/165216/paul-mcguire
        in answer to the question formulated by:
            Thales MG, http://stackoverflow.com/users/2708711/thales-mg
        the code is licensed under 'CC-WIKI'.
        (see: http://blog.stackoverflow.com/2009/06/attribution-required/)

    """

    LPAR, RPAR = map(Suppress, "()")
    integer = Word(nums)

    # add parse action to convert integers to ints, to support doing addition
    # and multiplication at parse time
    integer.setParseAction(lambda t: int(t[0]))

    # element = Word(alphas.upper(), alphas.lower())
    # or if you want to be more specific, use this Regex
    element = Regex(
        r"A[cglmrstu]|B[aehikr]?|C[adeflmorsu]?|D[bsy]|E[rsu]|F[emr]?|"
        "G[ade]|H[efgos]?|I[nr]?|Kr?|L[airu]|M[dgnot]|N[abdeiop]?|"
        "Os?|P[abdmortu]?|R[abefghnu]|S[bcegimnr]?|T[abcehilm]|"
        "Uu[bhopqst]|U|V|W|Xe|Yb?|Z[nr]")

    # forward declare 'formula' so it can be used in definition of 'term'
    formula = Forward()

    term = Group((element | Group(LPAR + formula + RPAR)("subgroup")) +
                 Optional(integer, default=1)("mult"))

    # define contents of a formula as one or more terms
    formula << OneOrMore(term)

    # add parse actions for parse-time processing

    # parse action to multiply out subgroups
    def multiplyContents(tokens):
        t = tokens[0]
        # if these tokens contain a subgroup, then use multiplier to
        # extend counts of all elements in the subgroup
        if t.subgroup:
            mult = t.mult
            for term in t.subgroup:
                term[1] *= mult
            return t.subgroup
    term.setParseAction(multiplyContents)

    # add parse action to sum up multiple references to the same element
    def sumByElement(tokens):
        elementsList = [t[0] for t in tokens]

        # construct set to see if there are duplicates
        duplicates = len(elementsList) > len(set(elementsList))

        # if there are duplicate element names, sum up by element and
        # return a new nested ParseResults
        if duplicates:
            ctr = defaultdict(int)
            for t in tokens:
                ctr[t[0]] += t[1]
            return ParseResults([ParseResults([k, v]) for k, v in ctr.items()])
    formula.setParseAction(sumByElement)

    return formula

_parser = _get_formula_parser()

symbols = (
    'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al',
    'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe',
    'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr',
    'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn',
    'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm',
    'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W',
    'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn',
    'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf',
    'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds',
    'Rg', 'Cn', 'Uut', 'Fl', 'Uup', 'Lv', 'Uus', 'Uuo'
)

names = (
    'Hydrogen', 'Helium', 'Lithium', 'Beryllium',
    'Boron', 'Carbon', 'Nitrogen', 'Oxygen', 'Fluorine', 'Neon', 'Sodium',
    'Magnesium', 'Aluminium', 'Silicon', 'Phosphorus', 'Sulfur',
    'Chlorine', 'Argon', 'Potassium', 'Calcium', 'Scandium', 'Titanium',
    'Vanadium', 'Chromium', 'Manganese', 'Iron', 'Cobalt', 'Nickel',
    'Copper', 'Zinc', 'Gallium', 'Germanium', 'Arsenic', 'Selenium',
    'Bromine', 'Krypton', 'Rubidium', 'Strontium', 'Yttrium', 'Zirconium',
    'Niobium', 'Molybdenum', 'Technetium', 'Ruthenium', 'Rhodium',
    'Palladium', 'Silver', 'Cadmium', 'Indium', 'Tin', 'Antimony',
    'Tellurium', 'Iodine', 'Xenon', 'Caesium', 'Barium', 'Lanthanum',
    'Cerium', 'Praseodymium', 'Neodymium', 'Promethium', 'Samarium',
    'Europium', 'Gadolinium', 'Terbium', 'Dysprosium', 'Holmium',
    'Erbium', 'Thulium', 'Ytterbium', 'Lutetium', 'Hafnium', 'Tantalum',
    'Tungsten', 'Rhenium', 'Osmium', 'Iridium', 'Platinum', 'Gold',
    'Mercury', 'Thallium', 'Lead', 'Bismuth', 'Polonium', 'Astatine',
    'Radon', 'Francium', 'Radium', 'Actinium', 'Thorium', 'Protactinium',
    'Uranium', 'Neptunium', 'Plutonium', 'Americium', 'Curium',
    'Berkelium', 'Californium', 'Einsteinium', 'Fermium', 'Mendelevium',
    'Nobelium', 'Lawrencium', 'Rutherfordium', 'Dubnium', 'Seaborgium',
    'Bohrium', 'Hassium', 'Meitnerium', 'Darmstadtium', 'Roentgenium',
    'Copernicium', '(Ununtrium)', 'Flerovium', '(Ununpentium)',
    'Livermorium', '(Ununseptium)', '(Ununoctium)'
)

lower_names = tuple(n.lower().lstrip('(').rstrip(')') for n in names)


def atomic_number(name):
    try:
        return symbols.index(name) + 1
    except ValueError:
        return lower_names.index(name.lower()) + 1

# The data in '_relative_atomic_masses' is licensed under the CC-SA license
# https://en.wikipedia.org/w/index.php?title=List_of_elements&oldid=700476748
_relative_atomic_masses = (
    "1.008 4.002602(2) 6.94 9.0121831(5) 10.81 12.011 14.007 15.999"
    " 18.998403163(6) 20.1797(6) 22.98976928(2) 24.305 26.9815385(7) 28.085"
    " 30.973761998(5) 32.06 35.45 39.948(1) 39.0983(1) 40.078(4)"
    " 44.955908(5) 47.867(1) 50.9415(1) 51.9961(6) 54.938044(3) 55.845(2)"
    " 58.933194(4) 58.6934(4) 63.546(3) 65.38(2) 69.723(1) 72.630(8)"
    " 74.921595(6) 78.971(8) 79.904 83.798(2) 85.4678(3) 87.62(1)"
    " 88.90584(2) 91.224(2) 92.90637(2) 95.95(1) [98] 101.07(2) 102.90550(2)"
    " 106.42(1) 107.8682(2) 112.414(4) 114.818(1) 118.710(7) 121.760(1)"
    " 127.60(3) 126.90447(3) 131.293(6) 132.90545196(6) 137.327(7)"
    " 138.90547(7) 140.116(1) 140.90766(2) 144.242(3) [145] 150.36(2)"
    " 151.964(1) 157.25(3) 158.92535(2) 162.500(1) 164.93033(2) 167.259(3)"
    " 168.93422(2) 173.045(10) 174.9668(1) 178.49(2) 180.94788(2) 183.84(1)"
    " 186.207(1) 190.23(3) 192.217(3) 195.084(9) 196.966569(5) 200.592(3)"
    " 204.38 207.2(1) 208.98040(1) [209] [210] [222] [223] [226] [227]"
    " 232.0377(4) 231.03588(2) 238.02891(3) [237] [244] [243] [247] [247]"
    " [251] [252] [257] [258] [259] [266] [267] [268] [269] [270] [269]"
    " [278] [281] [282] [285] [286] [289] [289] [293] [294] [294]"
)


def _get_relative_atomic_masses():
    for mass in _relative_atomic_masses.split():
        if mass.startswith('[') and mass.endswith(']'):
            yield float(mass[1:-1])
        elif '(' in mass:
            yield float(mass.split('(')[0])
        else:
            yield(float(mass))

relative_atomic_masses = tuple(_get_relative_atomic_masses())


def mass_from_composition(composition):
    """ Calculates molecular mass from atomic weights

    Parameters
    ----------
    composition: dict
        Dictionary mapping int (atomic number) to int (coefficient)

    Returns
    -------
    float
        molecular weight in atomic mass units


    Notes
    -----
    Atomic number 0 denotes charge or "net electron defficiency"

    Examples
    --------
    >>> '%.2f' % mass_from_composition({0: -1, 1: 1, 8: 1})
    '17.01'
    """
    mass = 0.0
    for k, v in composition.items():
        if k == 0:  # electron
            mass -= v*5.489e-4
        else:
            mass += v*relative_atomic_masses[k-1]
    return mass


def _get_charge(chgstr):

    if chgstr == '+':
        return 1
    elif chgstr == '-':
        return -1

    for token, anti, sign in zip('+-', '-+', (1, -1)):
        if token in chgstr:
            if anti in chgstr:
                raise ValueError("Invalid charge description (+ & - present)")
            before, after = chgstr.split(token)
            if len(before) > 0 and len(after) > 0:
                raise ValueError("Values both before and after charge token")
            if len(before) > 0:
                # will_be_missing_in='0.4.0'
                warnings.warn("'Fe/3+' deprecated, use e.g. 'Fe+3'",
                              ChemPyDeprecationWarning, stacklevel=3)
                return sign * int(1 if before == '' else before)
            if len(after) > 0:
                return sign * int(1 if after == '' else after)
    raise ValueError("Invalid charge description (+ or - missing)")


def _formula_to_parts(formula, prefixes, suffixes):
    # Drop prefixes and suffixes
    drop_pref, drop_suff = [], []
    for ign in prefixes:
        if formula.startswith(ign):
            drop_pref.append(ign)
            formula = formula[len(ign):]
    for ign in suffixes:
        if formula.endswith(ign):
            drop_suff.append(ign)
            formula = formula[:-len(ign)]

    # Extract charge
    if '/' in formula:
        # will_be_missing_in='0.4.0'
        warnings.warn("/ depr. (before 0.4.0): use 'Fe+3' over 'Fe/3+'",
                      ChemPyDeprecationWarning, stacklevel=3)
        parts = formula.split('/')

        if '+' in parts[0] or '-' in parts[0]:
            raise ValueError("Charge needs to be separated with a /")
        if parts[1] is not None:
            wo_pm = parts[1].replace('+', '').replace('-', '')
            if wo_pm != '' and not str.isdigit(wo_pm):
                raise ValueError("Non-digits in charge specifier")
        if len(parts) > 2:
            raise ValueError("At most one '/' allowed in formula")
    else:
        for token in '+-':
            if token in formula:
                if formula.count(token) > 1:
                    raise ValueError("Multiple tokens: %s" % token)
                parts = formula.split(token)
                parts[1] = token + parts[1]
                break
        else:
            parts = [formula, None]
    return parts + [tuple(drop_pref), tuple(drop_suff[::-1])]


def _parse_stoich(stoich):
    if stoich == 'e':  # special case, the electron is not an element
        return {}
    return {symbols.index(k)+1: n for k, n in _parser.parseString(stoich)}

_greek_letters = (
    'alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta', 'eta', 'theta',
    'iota', 'kappa', 'lambda', 'mu', 'nu', 'xi', 'o', 'pi', 'rho', 'sigma',
    'tau', 'upsilon', 'phi', 'chi', 'psi', 'omega'
)

_latex_mapping = {k + '-': '\\' + k + '-' for k in _greek_letters}
_latex_mapping['epsilon-'] = '\\varepsilon-'
_latex_mapping['.'] = '^\\bullet '


def to_composition(formula,
                   prefixes=None,
                   suffixes=('(s)', '(l)', '(g)', '(aq)')):
    """ Parse composition of formula representing a chemical formula

    Composition is represented as a dict mapping int -> int (atomic
    number -> multiplicity). "Atomic number" 0 represents net charge.

    Parameters
    ----------
    formula: str
        Chemical formula, e.g. 'H2O', 'Fe+3', 'Cl-'
    prefixes: iterable strings
        Prefixes to ignore, e.g. ('.', 'alpha-')
    suffixes: tuple of strings
        Suffixes to ignore, e.g. ('(g)', '(s)')

    Examples
    --------
    >>> to_composition('NH4+') == {0: 1, 1: 4, 7: 1}
    True
    >>> to_composition('.NHO-(aq)') == {0: -1, 1: 1, 7: 1, 8: 1}
    True

    """
    if prefixes is None:
        prefixes = _latex_mapping.keys()
    parts = _formula_to_parts(formula, prefixes, suffixes)
    comp = _parse_stoich(parts[0])
    if parts[1] is not None:
        comp[0] = _get_charge(parts[1])
    return comp


def _subs(string, patterns):
    for patt, repl in patterns.items():
        string = string.replace(patt, repl)
    return string


def to_latex(formula,
             prefixes=None,
             suffixes=('(s)', '(l)', '(g)', '(aq)'),
             intercept=None):
    r""" Convert formula string to latex representation

    Parameters
    ----------
    formula: str
        Chemical formula, e.g. 'H2O', 'Fe/3+', 'Cl-'
    prefixes: dict
        Prefix transofmrations, default: greek letters and .
    suffixes: tuple of strings
        Suffixes to keep, e.g. ('(g)', '(s)')

    Examples
    --------
    >>> to_latex('NH4+')
    'NH_{4}^{+}'
    >>> to_latex('Fe(CN)6+2')
    'Fe(CN)_{6}^{2+}'
    >>> to_latex('Fe(CN)6+2(aq)')
    'Fe(CN)_{6}^{2+}(aq)'
    >>> to_latex('.NHO-(aq)')
    '^\\bullet NHO^{-}(aq)'
    >>> to_latex('alpha-FeOOH(s)')
    '\\alpha-FeOOH(s)'

    """
    if prefixes is None:
        prefixes = _latex_mapping
    parts = _formula_to_parts(formula, prefixes.keys(), suffixes)

    string = re.sub(r'([0-9]+)', r'_{\1}', parts[0])
    if parts[1] is not None:
        chg = _get_charge(parts[1])
        if chg < 0:
            token = '-' if chg == -1 else '%d-' % -chg
        if chg > 0:
            token = '+' if chg == 1 else '%d+' % chg
        string += '^{%s}' % token
    if len(parts) > 4:
        raise ValueError("Incorrect formula")
    pre_str = ''.join(map(lambda x: _subs(x, prefixes), parts[2]))
    return pre_str + string + ''.join(parts[3])


def _parse_multiplicity(strings, substance_keys=None):
    """
    Examples
    --------
    >>> _parse_multiplicity(['2 H2O2', 'O2']) == {'H2O2': 2, 'O2': 1}
    True
    >>> _parse_multiplicity(['2 * H2O2', 'O2']) == {'H2O2': 2, 'O2': 1}
    True

    """
    result = {}
    for items in [re.split(' \* | ', s) for s in strings]:
        if len(items) == 1:
            result[items[0]] = 1
        elif len(items) == 2:
            result[items[1]] = int(items[0])
        else:
            raise ValueError("To many parts in substring")
    if substance_keys is not None:
        for k in result:
            if k not in substance_keys:
                raise ValueError("Unkown substance_key: %s" % k)
    return result


def to_reaction(line, substance_keys, token, Cls, globals_=None):
    """ Parses a string into a Reaction object and substances

    Reac1 + 2 Reac2 + (2 Reac1) -> Prod1 + Prod2; 10**3.7; ref='doi:12/ab'
    Reac1 = Prod1; 2.1;

    Parameters
    ----------
    line: str
        string representation to be parsed
    substance_keys: iterable of strings
        Allowed names, e.g. ('H2O', 'H+', 'OH-')
    token : str
        delimiter token between reactant and product side
    Cls : class
        e.g. subclass of Reaction
    globals_: dict (optional)
        Globals passed on to :func:`eval`, when ``None``:
        `chempy.units.default_units` is used with 'chempy'
        and 'default_units' extra entries.

    Notes
    -----
    This function calls :func:`eval`, hence there are severe security concerns
    with running this on untrusted data.

    """
    # TODO: add handling of units.
    if globals_ is None:
        import chempy
        from chempy.units import default_units
        globals_ = {'chempy': chempy, 'default_units': default_units}
        globals_.update(default_units.as_dict())

    try:
        stoich, param, kwargs = map(str.strip, line.rstrip('\n').split(';'))
    except ValueError:
        stoich, param = map(str.strip, line.rstrip('\n').split(';'))
        kwargs = {}
    else:
        kwargs = eval('dict('+kwargs+')', globals_)

    param = eval(param, globals_)

    if token not in stoich:
        raise ValueError("Missing token: %s" % token)

    reac_prod = [[y.strip() for y in x.split(' + ')] for
                 x in stoich.split(token)]

    act, inact = [], []
    for side in reac_prod:
        if side[-1].startswith('('):
            if not side[-1].endswith(')'):
                raise ValueError("Bad format (missing closing paren)")
            inact.append(_parse_multiplicity(side[-1][1:-1].split(' + '), substance_keys))
            act.append(_parse_multiplicity(side[:-1], substance_keys))
        else:
            inact.append({})
            act.append(_parse_multiplicity(side, substance_keys))

    # stoich coeff -> dict
    return Cls(act[0], act[1], param, inact_reac=inact[0],
               inact_prod=inact[1], **kwargs)



# -*- coding: utf-8 -*-

from __future__ import division

from itertools import chain
from operator import itemgetter
from collections import OrderedDict

import numpy as np

from .arrhenius import arrhenius_equation
from .util.arithmeticdict import ArithmeticDict
from .util.parsing import (
    to_composition, mass_from_composition, to_latex, to_reaction
)
from .util.pyutil import defaultnamedtuple, deprecated
from .units import to_unitless, default_constants, default_units


class Substance(object):
    """ Class representing a chemical substance

    Parameters
    ----------
    name: str
    charge: int (optional, default: None)
        will be stored in composition[0], prefer composition when possible
    latex_name: str
    composition: dict or None (default)
        dict (int -> number) e.g. {atomic number: count}, zero has special
        meaning (net charge)
    other_properties: dict
        free form dictionary. Could be simple such as ``{'mp': 0, 'bp': 100}``
        or considerably more involved, e.g.: ``{'diffusion_coefficient': {\
 'water': lambda T: 2.1*m**2/s/K*(T - 273.15*K)}}``

    Attributes
    ----------
    mass
        maps to other_properties, and when unavailable looks for formula.mass
    attrs
        a tuple of attribute names for serialization

    Examples
    --------
    >>> ammonium = Substance('NH4+', 1, 'NH_4^+', composition={14:1, 1: 4},
    ...     other_properties={'mass': 18.0385, 'pKa': 9.24})
    >>> ammonium.name
    'NH4+'
    >>> ammonium.composition  # note that charge was inserted as composition[0]
    {0: 1, 1: 4, 14: 1}
    >>> ammonium.other_properties['mass']
    18.0385
    >>> ammonium.other_properties['pKa']
    9.24
    >>> ammonium.mass  # mass is a special case (also attribute)
    18.0385
    >>> ammonium.pKa
    Traceback (most recent call last):
        ...
    AttributeError: 'Substance' object has no attribute 'pKa'

    """

    attrs = ('name', 'latex_name', 'composition', 'other_properties')

    @property
    def charge(self):
        """ Convenience property for accessing ``composition[0]`` """
        return self.composition.get(0, 0)  # electron (net) deficiency

    @property
    def mass(self):
        """ Convenience property for accessing ``other_properties['mass']``

        when ``other_properties['mass']`` is missing the mass is calculated
        from the :attr:`composition` using
        :func:`chempy.util.parsing.mass_from_composition`.
        """
        try:
            return self.other_properties['mass']
        except KeyError:
            if self.composition is not None:
                return mass_from_composition(self.composition)

    def __init__(self, name=None, charge=None, latex_name=None,
                 composition=None, other_properties=None):
        self.name = name
        self.latex_name = latex_name
        self.composition = composition or {}

        if 0 in self.composition:
            if charge is not None:
                raise KeyError("Cannot give both charge and composition[0]")
        else:
            if charge is not None:
                self.composition[0] = charge
        self.other_properties = other_properties or {}

    @classmethod
    def from_formula(cls, formula, **kwargs):
        """ Creates a :class:`Substance` instance from its formula

        Parameters
        ----------
        formula: str
            e.g. 'Na+', 'H2O', 'Fe(CN)6-4'
        \*\*kwargs:
            keyword arguments passed on to `.Substance`

        Examples
        --------
        >>> NH3 = Substance.from_formula('NH3')
        >>> NH3.composition == {1: 3, 7: 1}
        True
        >>> '%.2f' % NH3.mass
        '17.03'
        >>> NH3.charge
        0
        >>> NH3.latex_name
        'NH_{3}'

        """

        return cls(formula, latex_name=to_latex(formula),
                   composition=to_composition(formula),
                   **kwargs)

    def __repr__(self):
        kw = ['name=' + self.name + ', ...']  # Too verbose to print all
        return "<{}({})>".format(self.__class__.__name__, ','.join(kw))

    def __str__(self):
        return str(self.name)

    @staticmethod
    def composition_keys(substance_iter):
        """ Occuring :attr:`composition` keys among a series of substances """
        keys = set()
        for s in substance_iter:
            for k in s.composition.keys():
                keys.add(k)
        return sorted(keys)


class Species(Substance):
    """ Substance belonging to a phase

    Species extends :class:`Substance` with the new attribute :attr:`phase_idx`

    Attributes
    ----------
    phase_idx: int
        Index of the phase (default is 0)
    """
    def __init__(self, *args, **kwargs):
        phase_idx = kwargs.pop('phase_idx', 0)
        super(Species, self).__init__(*args, **kwargs)
        self.phase_idx = phase_idx

    @property
    @deprecated(last_supported_version='0.3.0', will_be_missing_in='0.4.0')
    def precipitate(self):
        """ deprecated attribute, provided for compatibility for now """
        return self.phase_idx > 0

    @classmethod
    def from_formula(cls, formula, phases=('(s)', '(l)', '(g)'),
                     default_phase_idx=0, **kwargs):
        """ Create a :class:`Species` instance from its formula

        Analogous to :meth:`Substance.from_formula` but with the addition that
        phase_idx is determined from the formula (and a mapping provided by
        ``phases``)

        Parameters
        ----------
        formula: str
            e.g. 'H2O', 'NaCl(s)', 'CO2(aq)', 'CO2(g)'
        phases: iterable of str or dict mapping str -> int
            if not in \*\*kwargs, ``phase_idx`` is determined from the suffix
            of ``formula`` where the suffixes is mapped from phases:
                if ``phases`` is a dictionary:
                    ``phase_idx = phases[suffix]``
                else:
                    ``phase_idx = phases.index(suffix) + 1``
            and if suffixes is missing in phases phase_idx is taken to be 0
        default_phase_idx: int or None (default: 0)
            If ``default_phase_idx`` is ``None``, ``ValueError`` is raised for
                unkown suffixes.
            Else ``default_phase_idx`` is used as ``phase_idx`` in those cases.
        \*\*kwargs:
            Keyword arguments passed on.

        Examples
        --------
        >>> water = Species.from_formula('H2O')
        >>> water.phase_idx
        0
        >>> NaCl = Species.from_formula('NaCl(s)')
        >>> NaCl.phase_idx
        1
        >>> Hg_l = Species.from_formula('Hg(l)')
        >>> Hg_l.phase_idx
        2
        >>> CO2g = Species.from_formula('CO2(g)')
        >>> CO2g.phase_idx
        3
        >>> CO2aq = Species.from_formula('CO2(aq)', default_phase_idx=None)
        Traceback (most recent call last):
            ...
        ValueError: Could not determine phase_idx
        >>> CO2aq = Species.from_formula('CO2(aq)')
        >>> CO2aq.phase_idx
        0
        >>> CO2aq = Species.from_formula('CO2(aq)', ['(aq)'],
        ...     default_phase_idx=None)
        >>> CO2aq.phase_idx
        1
        >>> Species.from_formula('CO2(aq)', {'(aq)': 0}, None).phase_idx
        0


        Raises
        ------
        ValueError:
            if ``default_phase_idx`` is ``None`` and no suffix found in phases

        """
        if 'phase_idx' in kwargs:
            p_i = kwargs.pop('phase_idx')
        else:
            p_i = None
            if isinstance(phases, dict):
                for k, v in phases.items():
                    if formula.endswith(k):
                        p_i = v
                        break
            else:
                for idx, phase in enumerate(phases):
                    if formula.endswith(phase):
                        p_i = idx + 1
                        break
            if p_i is None:
                if default_phase_idx is None:
                    raise ValueError("Could not determine phase_idx")
                else:
                    p_i = default_phase_idx
        return super(Species, cls).from_formula(
            formula, phase_idx=p_i, **kwargs)


@deprecated(last_supported_version='0.3.0',
            will_be_missing_in='0.4.0', use_instead=Species)
class Solute(Substance):
    """ [DEPRECATED] Use `.Species` instead

    Counter-intuitive to its name Solute has an additional
    property 'precipitate'

    """

    def __init__(self, *args, **kwargs):
        precipitate = kwargs.pop('precipitate', False)
        Substance.__init__(self, *args, **kwargs)
        self.precipitate = precipitate

    @classmethod
    def from_formula(cls, formula, **kwargs):
        if formula.endswith('(s)'):
            kwargs['precipitate'] = True
        return cls(formula, latex_name=to_latex(formula),
                   composition=to_composition(formula),
                   **kwargs)


class Reaction(object):
    """ Class representing a chemical reaction

    Consider for example:

        A + R --> A + P; r = k*A*R

    this would be represented as ``Reaction({'A': 1, 'R': 1},
    {'A': 1, 'P': 1}, param=k)``. Some reactions have a larger
    stoichiometric coefficient than what appears in the rate
    expression, e.g.:

        5*C1 + C2 --> B; r = k*C1*C2

    this can be represented as ``Reaction({'C1': 1, 'C2': 1},
    {'B': 1}, inact_reac={'C1': 4}, param=k)``.

    The rate constant information in ``param`` may be callable (with a single
    argument representing the state, e.g. temperature)

    Additional data may be stored in the ``other_properties`` dict.


    Parameters
    ----------
    reac: dict (str -> int)
    prod: dict (str -> int)
    param: float or callable
    inact_reac: dict (optional)
    inact_prod: dict (optional)
    name: str (optional)
    k: deprecated (alias for param)
    ref: object
        reference
    other_properties: dict (optional)
    """

    str_arrow = '->'
    latex_arrow = r'\rightarrow'
    param_char = 'k'  # convention

    def __init__(self, reac, prod, param=None, inact_reac=None,
                 inact_prod=None, name=None, ref=None,
                 other_properties=None):
        self.reac = reac
        self.prod = prod
        self.param = param
        self.inact_reac = inact_reac or {}
        self.inact_prod = inact_prod or {}
        self.name = name
        self.ref = ref
        self.other_properties = other_properties or {}

    @classmethod
    def from_string(cls, string, substance_keys, globals_=None):
        """ Parses a string into an instance

        Parameters
        ----------
        string : str
            string representation of the reaction
        substance_keys : iterable of strings or string
        globals_ : dict (optional)
            dict for eval for (default: None -> {'chempy': chempy})

        Examples
        --------
        >>> r = Reaction.from_string("H2O -> H+ + OH-; 1e-4", 'H2O H+ OH-')
        >>> r.reac == {'H2O': 1} and r.prod == {'H+': 1, 'OH-': 1}
        True

        Notes
        -----
        :func:`chempy.util.parsing.to_reaction` is used which in turn calls
        eval which is a severe security concern for untrusted input.

        """
        if isinstance(substance_keys, str):
            if ' ' in substance_keys:
                substance_keys = substance_keys.split()
        return to_reaction(string, substance_keys, cls.str_arrow, cls, globals_)

    def __eq__(lhs, rhs):
        if not isinstance(lhs, Reaction) or not isinstance(rhs, Reaction):
            return NotImplemented
        for attr in ['reac', 'prod', 'param', 'inact_reac', 'inact_prod']:
            if getattr(lhs, attr) != getattr(rhs, attr):
                return False
        return True

    def order(self):
        """ Sum of (active) reactant stoichiometries """
        return sum(self.reac.values())

    def net_stoich(self, substance_keys):
        """ Per substance net stoichiometry tuple (active & inactive) """
        return tuple(self.prod.get(k, 0) -
                     self.reac.get(k, 0) +
                     self.inact_prod.get(k, 0) -
                     self.inact_reac.get(k, 0) for k in substance_keys)

    def all_reac_stoich(self, substances):
        """ Per substance reactant stoichiometry tuple (active & inactive) """
        return tuple(self.reac.get(k, 0) + self.inact_reac.get(k, 0)
                     for k in substances)

    def all_prod_stoich(self, substances):
        """ Per substance product stoichiometry tuple (active & inactive) """
        return tuple(self.prod.get(k, 0) + self.inact_prod.get(k, 0)
                     for k in substances)

    def _xprecipitate_stoich(self, substances, xor):
        return tuple((
            0 if xor ^ v.precipitate else
            self.prod.get(k, 0) + self.inact_prod.get(k, 0) -
            self.reac.get(k, 0) - self.inact_reac.get(k, 0)
        ) for k, v in substances.items())

    def precipitate_stoich(self, substances):
        """ Only stoichiometry of precipitates """
        net = self._xprecipitate_stoich(substances, True)
        found1 = -1
        for idx in range(len(net)):
            if net[idx] != 0:
                if found1 == -1:
                    found1 = idx
                else:
                    raise NotImplementedError("Only one precipitate assumed.")
        return net, net[idx], idx

    def non_precipitate_stoich(self, substances):
        """ Only stoichiometry of non-precipitates """
        return self._xprecipitate_stoich(substances, False)

    def has_precipitates(self, substances):
        for s_name in chain(self.reac.keys(), self.prod.keys(),
                            self.inact_reac.keys(),
                            self.inact_prod.keys()):
            if substances[s_name].precipitate:
                return True
        return False

    def _get_str_parts(self, name_attr, arrow_attr, substances):
        reac, prod, i_reac, i_prod = [[
            ((str(v)+' ') if v > 1 else '') + str(getattr(
                substances[k], name_attr, k))
            for k, v in filter(itemgetter(1), d.items())
        ] for d in (self.reac, self.prod, self.inact_reac,
                    self.inact_prod)]
        r_str = " + ".join(sorted(reac))
        ir_str = ' (+ ' + " + ".join(sorted(i_reac)) + ')' if len(i_reac) > 0 else ''
        arrow_str = getattr(self, arrow_attr)
        p_str = " + ".join(sorted(prod))
        ip_str = ' (+ ' + " + ".join(sorted(i_prod)) + ')' if len(i_prod) > 0 else ''
        return r_str, ir_str, arrow_str, p_str, ip_str

    def _get_str(self, *args):
        return "{}{} {} {}{}".format(*self._get_str_parts(*args))

    def __str__(self):
        try:
            str_param = '%.3g %s' % (self.param, self.param.dimensionality)
        except AttributeError:
            try:
                str_param = '%.3g' % self.param
            except TypeError:
                str_param = str(self.param)
        s = '; ' + str_param
        return self._get_str('name', 'str_arrow', {
            k: k for k in chain(self.reac.keys(), self.prod.keys(),
                                self.inact_reac.keys(),
                                self.inact_prod.keys())
        }) + s

    def latex(self, substances):
        """ Returns a LaTeX representation of the reaction """
        return self._get_str('latex_name', 'latex_arrow', substances)

    def _violation(self, substances, attr):
        net = 0.0
        for substance, coeff in zip(substances.values(),
                                    self.net_stoich(substances.keys())):
            net += getattr(substance, attr) * coeff
        return net

    def mass_balance_violation(self, substances):
        """ Net amount of mass produced

        Parameters
        ----------
        substances: dict

        Returns
        -------
        float: amount of net mass produced/consumed

        """
        return self._violation(substances, 'mass')

    def charge_neutrality_violation(self, substances):
        """ Net amount of charge produced

        Parameters
        ----------
        substances: dict

        Returns
        -------
        float: amount of net charge produced/consumed

        """
        return self._violation(substances, 'charge')

    def composition_violation(self, substances, composition_keys=None):
        """ Net amount of constituent produced

        If composition keys correspond to conserved entities e.g. atoms
        in chemical reactions, this function should return a list of zeros.
        """
        if composition_keys is None:
            composition_keys = Substance.composition_keys(substances.values())
        net = [0]*len(composition_keys)
        for substance, coeff in zip(substances.values(),
                                    self.net_stoich(substances.keys())):
            for idx, key in enumerate(composition_keys):
                net[idx] += substance.composition.get(key, 0) * coeff
        return net


def equilibrium_quotient(concs, stoich):
    """ Calculates the equilibrium quotient of an equilbrium

    Parameters
    ----------
    concs: array_like
        per substance concentration
    stoich: iterable of integers
        per substance stoichiometric coefficient

    Examples
    --------
    >>> '%.12g' % equilibrium_quotient([1.0, 1e-7, 1e-7], [-1, 1, 1])
    '1e-14'

    """
    if not hasattr(concs, 'ndim') or concs.ndim == 1:
        tot = 1
    else:
        tot = np.ones(concs.shape[0])
        concs = concs.T

    for nr, conc in zip(stoich, concs):
        tot *= conc**nr
    return tot


class ArrheniusRate(defaultnamedtuple('ArrheniusRate', 'A Ea ref', [None])):
    """ Kinetic data in the form of an Arrhenius parametrization

    Parameters
    ----------
    Ea: float
        activation energy
    A: float
        preexponential prefactor (Arrhenius type eq.)
    ref: object (default: None)
        arbitrary reference (e.g. string representing citation key)

    Examples
    --------
    >>> k = ArrheniusRate(1e13, 40e3)
    >>> '%.5g' % k(298.15)
    '9.8245e+05'

    """
    def __call__(self, T, constants=None, units=None, exp=None):
        """ See :func:`chempy.arrhenius.arrhenius_equation`. """
        return arrhenius_equation(self.A, self.Ea, T, constants=constants,
                                  units=units, exp=exp)

    @staticmethod
    def _fmt(arg, precision, tex):
        if tex:
            unit_str = arg.dimensionality.latex.strip('$')
        else:
            from quantities.markup import config
            attr = 'unicode' if config.use_unicode else 'string'
            unit_str = getattr(arg.dimensionality, attr)
        return precision.format(float(arg.magnitude)) + " " + unit_str

    def format(self, precision, tex=False):
        try:
            str_A = self._fmt(self.A, precision, tex)
            str_Ea = self._fmt(self.Ea, precision, tex)
        except:
            str_A = precision.format(self.A)
            str_Ea = precision.format(self.Ea)
        return str_A, str_Ea

    def equation_as_string(self, precision, tex=False):
        if tex:
            return r"{}\exp \left(\frac{{{}}}{{RT}} \right)".format(
                *self.format(precision, tex))
        else:
            return "{}*exp({}/(R*T))".format(*self.format(precision, tex))

    def __str__(self):
        return self.equation_as_string('{0:.5g}')


class ArrheniusRateWithUnits(ArrheniusRate):
    def __call__(self, T, constants=default_constants, units=default_units,
                 exp=None):
        """ See :func:`chempy.arrhenius.arrhenius_equation`. """
        return super(ArrheniusRateWithUnits, self).__call__(
            T, constants, units, exp)


class Equilibrium(Reaction):
    """ Represents an equilibrium reaction

    See :class:`Reaction` for parameters

    """
    str_arrow = '='
    latex_arrow = r'\rightleftharpoons'
    param_char = 'K'  # convention

    def as_reactions(self, state=None, kf=None, kb=None, units=None):
        """ Creates a forward and backward :class:`Reaction` pair """
        nb = sum(self.prod.values())
        nf = sum(self.reac.values())
        if units is None:
            if hasattr(kf, 'units') or hasattr(kb, 'units'):
                raise ValueError("units missing")
            c0 = 1
        else:
            c0 = 1*units.molar  # standard concentration IUPAC

        if kf is None:
            if kb is None:
                raise ValueError("Exactly one rate needs to be provided")
            kf = kb * self.K(state) * c0**(nb - nf)
        elif kb is None:
            kb = kf / (self.K(state) * c0**(nb - nf))
        else:
            raise ValueError("Exactly one rate needs to be provided")
        return (
            Reaction(self.reac, self.prod, kf, self.inact_reac,
                     self.inact_prod, ref=self.ref),
            Reaction(self.prod, self.reac, kb, self.inact_prod,
                     self.inact_reac, ref=self.ref)
        )

    def K(self, state=None):
        """ Return equilibrium quotient (possibly state dependent) """
        if callable(self.param):
            # For some cases a default might be obvious:
            # if state is None:
            #     raise ValueError("No state provided")
            return self.param(state)
        else:
            if state is not None:
                raise ValueError("state provided but param not callable")
            return self.param

    def Q(self, substances, concs):
        """ Calculates the equilibrium qoutient """
        stoich = self.non_precipitate_stoich(substances)
        return equilibrium_quotient(concs, stoich)

    def precipitate_factor(self, substances, sc_concs):
        factor = 1
        for r, n in self.reac.items():
            if r.precipitate:
                factor *= sc_concs[substances.index(r)]**-n
        for p, n in self.prod.items():
            if p.precipitate:
                factor *= sc_concs[substances.index(p)]**n
        return factor

    def dimensionality(self):
        result = 0
        for r, n in self.reac.items():
            if r.precipitate:
                continue
            result -= n
        for p, n in self.prod.items():
            if p.precipitate:
                continue
            result += n
        return result

    def __rmul__(lhs, rhs):  # This works on both Py2 and Py3
        if not isinstance(rhs, int) or not isinstance(lhs, Equilibrium):
            return NotImplemented
        return Equilibrium(dict(rhs*ArithmeticDict(int, lhs.reac)),
                           dict(rhs*ArithmeticDict(int, lhs.prod)),
                           lhs.param**rhs)

    def __add__(lhs, rhs):
        keys = set()
        for key in chain(lhs.reac.keys(), lhs.prod.keys(),
                         rhs.reac.keys(), rhs.prod.keys()):
            keys.add(key)
        reac, prod = {}, {}
        for key in keys:
            n = (lhs.prod.get(key, 0) - lhs.reac.get(key, 0) +
                 rhs.prod.get(key, 0) - rhs.reac.get(key, 0))
            if n < 0:
                reac[key] = -n
            elif n > 0:
                prod[key] = n
            else:
                pass  # n == 0
        return Equilibrium(reac, prod, lhs.param * rhs.param)

    def __sub__(lhs, rhs):
        return lhs + -1*rhs


class ReactionSystem(object):
    """
    Collection of reactions forming a system (model).

    Parameters
    ----------
    rxns: sequence
         sequence of :py:class:`Reaction` instances
    substances: OrderedDict
         mapping str -> Substance instances
    name: string (optional)
         Name of ReactionSystem (e.g. model name / citation key)
    check_balance: bool (default: None)
        if None => True if all substances has composition attribute.

    Attributes
    ----------
    rxns: list of objects
        sequence of :class:`Reaction` instances
    substances: OrderedDict
        mapping substance name to substance index
    ns: int
        number of substances
    nr: int
        number of reactions

    Examples
    --------
    >>> from chempy import Reaction
    >>> r1 = Reaction({'R1': 1}, {'P1': 1}, 42.0)
    >>> rsys = ReactionSystem([r1], 'R1 P1')
    >>> rsys.as_per_substance_array({'R1': 2, 'P1': 3})
    array([ 2.,  3.])

    """

    def __init__(self, rxns, substances, name=None, check_balance=None):
        self.rxns = rxns
        if isinstance(substances, OrderedDict):
            self.substances = substances
        elif isinstance(substances, str):
            if ' ' in substances:
                substances = substances.split()
            self.substances = OrderedDict([
                (s, Substance(s)) for s in substances])
        else:
            try:
                self.substances = OrderedDict([(s.name, s) for s in substances])
            except:
                self.substances = OrderedDict(substances)
        self._sanity_check()
        self.name = name
        if check_balance is None:
            for subst in self.substances.values():
                if subst.composition is None:
                    check_balance = False
                    break
            else:
                check_balance = True
        if check_balance:
            self._balance_check()

    def _balance_check(self):
        for rxn in self.rxns:
            for net in rxn.composition_violation(self.substances):
                if net != 0:
                    raise ValueError("Reaction not balanced: %s" % str(rxn))

    def _sanity_check(self):
        for rxn in self.rxns:
            for key in chain(rxn.reac, rxn.prod, rxn.inact_reac,
                             rxn.inact_prod):
                if key not in self.substances:
                    raise ValueError("Unkown substance: %s" % key)

    def substance_names(self):
        """ Returns a tuple of the substances' names """
        return tuple(substance.name for substance in self.substances.values())

    @property
    def nr(self):
        """ Number of reactions """
        return len(self.rxns)

    @property
    def ns(self):
        """ Number of substances """
        return len(self.substances)

    def params(self):
        """ Returns list of per reaction ``param`` value """
        return [rxn.param for rxn in self.rxns]

    def as_per_substance_array(self, cont, dtype=np.float64, unit=None):
        """ Turns a dict into an ordered array """
        if unit is not None:
            cont = to_unitless(cont, unit)
        if isinstance(cont, np.ndarray):
            pass
        elif isinstance(cont, dict):
            substance_keys = self.substances.keys()
            for k in cont:
                if k not in substance_keys:
                    raise KeyError("Unkown substance key: %s" % k)
            cont = [cont[k] for k in substance_keys]

        cont = np.asarray(cont, dtype=dtype)
        if cont.shape[-1] != self.ns:
            raise ValueError("Incorrect size")
        return cont*(unit if unit is not None else 1)

    def as_substance_index(self, sbstnc):
        """ Returns the index of a Substance in the system"""
        if isinstance(sbstnc, int):
            return sbstnc
        else:
            return list(self.substances.keys()).index(sbstnc)

    def _stoichs(self, attr):
        # dtype: see https://github.com/sympy/sympy/issues/10295
        return np.array([(getattr(eq, attr)(self.substances.keys())) for
                         eq in self.rxns],
                        dtype=object)

    def net_stoichs(self):
        return self._stoichs('net_stoich')

    def all_reac_stoichs(self):
        return self._stoichs('all_reac_stoich')

    def all_prod_stoichs(self):
        return self._stoichs('all_prod_stoich')

    def stoichs(self, non_precip_rids=()):
        """ Conditional stoichiometries depending on precipitation status """
        # dtype: see https://github.com/sympy/sympy/issues/10295
        return np.array([(
            -np.array(eq.precipitate_stoich(self.substances)[0]) if idx
            in non_precip_rids else
            eq.non_precipitate_stoich(self.substances)
        ) for idx, eq in enumerate(self.rxns)],
                        dtype=object)

    def obeys_mass_balance(self):
        """ Returns True if all reactions obeys mass balance, else False. """
        for rxn in self.rxns:
            if rxn.mass_balance_violation(self.substances) != 0:
                return False
        return True

    def obeys_charge_neutrality(self):
        """ Returns False if any reaction violate charge neutrality. """
        for rxn in self.rxns:
            if rxn.charge_neutrality_violation(self.substances) != 0:
                return False
        return True



# -*- coding: utf-8 -*-
"""
This module contains functions for formulating systems of Ordinary Differential
Equations (ODE-systems) which may be integrated numerically to model temporal
evolution of concentrations in reaction systems.
"""
from __future__ import (absolute_import, division, print_function)

from chempy.units import get_derived_unit, to_unitless


def law_of_mass_action_rates(conc, rsys, k=None, state=None):
    """ Returns a generator of reaction rate expressions

    Rates according to the law of mass action (:attr:`rsys.inact_reac` ignored)
    from a :class:`ReactionSystem`.

    Parameters
    ----------
    conc: array_like
        concentrations (floats or symbolic objects)
    rsys: ReactionSystem instance
        See :class:`ReactionSystem`
    k: array_like (optional)
        to override rate parameters of the reactions
    state: object (optional)
        argument for reaction parameters

    Examples
    --------
    >>> from chempy import ReactionSystem, Reaction
    >>> line, keys = 'H2O -> H+ + OH- ; 1e-4', 'H2O H+ OH-'
    >>> rsys = ReactionSystem([Reaction.from_string(line, keys)], keys)
    >>> next(law_of_mass_action_rates([55.4, 1e-7, 1e-7], rsys))
    0.00554

    """
    def _eval_k(_k):
        is_func = callable(_k) and not _k.__class__.__name__ == 'Symbol'
        return _k(state) if is_func else _k
    for rxn_idx, rxn in enumerate(rsys.rxns):
        rate = 1
        for substance_key, coeff in rxn.reac.items():
            s_idx = rsys.as_substance_index(substance_key)
            rate *= conc[s_idx]**coeff
        if k is None:
            yield rate * _eval_k(rxn.param)
        else:
            yield rate * _eval_k(k[rxn_idx])


def dCdt(rsys, rates):
    """ Returns a list of the time derivatives of the concentrations

    Parameters
    ----------
    rsys: ReactionSystem instance
    rates: array_like
        rates (to be weighted by stoichiometries) of the reactions
        in ``rsys``

    Examples
    --------
    >>> from chempy import ReactionSystem, Reaction
    >>> line, keys = 'H2O -> H+ + OH- ; 1e-4', 'H2O H+ OH-'
    >>> rsys = ReactionSystem([Reaction.from_string(line, keys)], keys)
    >>> dCdt(rsys, [0.0054])
    [-0.0054, 0.0054, 0.0054]

    """
    f = [0]*rsys.ns
    net_stoichs = rsys.net_stoichs()
    for idx_s in range(rsys.ns):
        for idx_r in range(rsys.nr):
            f[idx_s] += net_stoichs[idx_r, idx_s]*rates[idx_r]
    return f


class _Always(object):
    __slots__ = ('value')

    def __init__(self, value):
        self.value = value

    def __getitem__(self, key):
        return self.value


def get_odesys(rsys, include_params=False, SymbolicSys=None,
               unit_registry=None, output_conc_unit=None,
               output_time_unit=None, state=None, **kwargs):
    """ Creates a :class:`pyneqsys.SymbolicSys` from a :class:`ReactionSystem`

    Parameters
    ----------
    rsys: ReactionSystem
    include_params: bool (default: False)
        whether rate constants should be included into the rate expressions or
        left as free parameters in the :class:`pyneqsys.SymbolicSys` instance.
    SymbolicSys: class (optional)
        default: :class:`pyneqsys.SymbolicSys`
    unit_registry: dict (optional)
        see :func:`chempy.units.get_derived_units`
    output_conc_unit: unit (Optional)
    output_time_unit: unit (Optional)
    state: object (optional)
        argument for reaction parameters
    \*\*kwargs:
        Keyword arguemnts pass on to `SymbolicSys`
    """
    if SymbolicSys is None:
        from pyodesys.symbolic import SymbolicSys

    if 'names' not in kwargs:
        kwargs['names'] = list(rsys.substances.keys())

    if unit_registry is not None:
        # We need to make rsys_params unitless and create
        # a post- & pre-processor for SymbolicSys
        time_unit = get_derived_unit(unit_registry, 'time')
        conc_unit = get_derived_unit(unit_registry, 'concentration')
        p_units = list(law_of_mass_action_rates(_Always(1/conc_unit), rsys,
                                                _Always(conc_unit/time_unit), state))
        rsys_params = [elem(state) if callable(elem) else to_unitless(elem, p_unit) for elem, p_unit
                       in zip(rsys.params(), p_units)]

        def pre_processor(x, y, p):
            return to_unitless(x, time_unit), to_unitless(y, conc_unit), [
                to_unitless(elem, p_unit) for elem, p_unit in zip(p, p_units)]

        def post_processor(x, y, p):
            time = x*time_unit
            if output_time_unit is not None:
                time = time.rescale(output_time_unit)
            conc = y*conc_unit
            if output_conc_unit is not None:
                conc = conc.rescale(output_conc_unit)
            return time, conc, [elem*p_unit for elem, p_unit
                                in zip(p, p_units)]
        kwargs['pre_processors'] = [pre_processor]
        kwargs['post_processors'] = [post_processor]
    else:
        rsys_params = rsys.params()

    def dydt(t, y, p):
        rates = list(law_of_mass_action_rates(
            y, rsys, rsys_params if include_params else p))
        return dCdt(rsys, rates)

    return SymbolicSys.from_callback(
        dydt, rsys.ns, 0 if include_params else rsys.nr, **kwargs)



import warnings

try:
    from numpy import exp as _exp
    from numpy import log as ln
    from numpy import any as _any
except ImportError:
    from math import exp as _exp
    from math import log as ln
    _any = any


def water_permittivity(T=None, P=None, units=None, U=None,
                       just_return_U=False, warn=True, exp=None):
    """
    Relative permittivity of water as function of temperature (K)
    and pressure (bar).

    Parameters
    ----------
    T: float
        Temperature (default: 298.15 Kelvin)
    P: float
        Pressure (default: 1 bar)
    units: object (optional)
        object with attributes: Kelvin, bar
    U: array_like (optional)
        9 parameters to the equation.
    just_return_U: bool (optional, default: False)
        Do not compute relative permittivity, just return the parameters ``U``.
    warn: bool (default: True)
        Emit UserWarning when outside temperature/pressure range.
    exp: callback (default: None)
        Custom callback for evaluating the exponential function

    Returns
    -------
    Relative permittivity of water (dielectric constant)

    References
    ----------
    Bradley, D.J.; Pitzer, K.S. `Thermodynamics of electrolytes. 12. Dielectric
        properties of water and Debye--Hueckel parameters to 350/sup
        0/C and 1 kbar`, J. Phys. Chem.; Journal Volume 83 (12)
        (1979), pp. 1599-1603,
        http://pubs.acs.org/doi/abs/10.1021/j100475a009
        DOI: 10.1021/j100475a009
    """
    if units is None:
        K = 1
        bar = 1
    else:
        K = units.kelvin
        bar = units.bar
    if T is None:
        T = 298.15*K
    if P is None:
        P = 1*bar
    if U is None:
        U = (3.4279e2,
             -5.0866e-3 / K,
             9.4690e-7 / K**2,
             -2.0525,
             3.1159e3*K,
             -1.8289e2*K,
             -8.0325e3*bar,
             4.2142e6*K*bar,
             2.1417/K*bar)
    if just_return_U:
        return U
    T0 = 273.15*K
    if warn:
        if _any(T < T0) or _any(T > T0 + 350*K):
            warnings.warn("Outside temperature range (0-350 degC)")
        else:
            if _any(T > T0 + 70*K):
                if _any(P > 2000*bar):
                    warnings.warn("Outside pressure range (2000 bar)")
                else:
                    if _any(P > 5000*bar):
                        warnings.warn("Outside pressure range (5000 bar)")
    if exp is None:
        exp = _exp
    B = U[6] + U[7]/T + U[8]*T
    C = U[3] + U[4]/(U[5] + T)
    eps1000 = U[0]*exp(U[1]*T + U[2]*T**2)
    return eps1000 + C*ln((B+P)/(B + 1000.0*bar))


# generated at doi2bib.org:
bibtex = """
@article{Bradley1979,
  doi = {10.1021/j100475a009},
  url = {http://dx.doi.org/10.1021/j100475a009},
  year  = {1979},
  month = {jun},
  publisher = {American Chemical Society ({ACS})},
  volume = {83},
  number = {12},
  pages = {1599--1603},
  author = {Daniel J. Bradley and Kenneth S. Pitzer},
  title = {Thermodynamics of electrolytes. 12. Dielectric properties of water
           and Debye-Hueckel parameters to 350.degree.C and 1 kbar},
  journal = {J. Phys. Chem.}
}
"""



import warnings

from chempy.units import allclose
from ..water_density_tanaka_2001 import water_density


def test_water_density():
    warnings.filterwarnings("error")
    assert abs(water_density(273.15 + 0) - 999.8395) < 0.004
    assert abs(water_density(273.15 + 4) - 999.9720) < 0.003
    assert abs(water_density(273.15 + 10) - 999.7026) < 0.0003
    assert abs(water_density(273.15 + 15) - 999.1026) < 0.0001
    assert abs(water_density(273.15 + 20) - 998.2071) < 0.0005
    assert abs(water_density(273.15 + 22) - 997.7735) < 0.0007
    assert abs(water_density(273.15 + 25) - 997.0479) < 0.0009
    assert abs(water_density(273.15 + 30) - 995.6502) < 0.0016
    assert abs(water_density(273.15 + 40) - 992.2) < 0.02

    try:
        water_density(1)
    except UserWarning:
        pass  # good warning raised
    else:
        raise
    warnings.resetwarnings()

    try:
        import quantities as pq
        import numpy as np
        unit = pq.kg/pq.m**3
        assert allclose(water_density(298.15*pq.K, units=pq),
                        997.047021671824*unit, atol=1e-8*unit)
        assert allclose(water_density(np.linspace(297, 299)*pq.K, units=pq),
                        997*unit, rtol=1e-3, atol=1e-3*unit)
    except ImportError:
        pass



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import math
from operator import attrgetter

import pytest

from ..units import default_units
from ..chemistry import (
    Substance, Species, Reaction, ReactionSystem, ArrheniusRate,
    ArrheniusRateWithUnits, Equilibrium
)


def test_Substance():
    s = Substance.from_formula('H+')
    assert s.composition == {0: 1, 1: 1}
    assert s.charge == 1
    assert abs(s.mass - 1.008) < 1e-3


def test_Substance__2():
    H2O = Substance(name='H2O',  charge=0, latex_name=r'$\mathrm{H_{2}O}$',
                    other_properties={'pKa': 14})
    OH_m = Substance(name='OH-',  charge=-1, latex_name=r'$\mathrm{OH^{-}}$')
    assert sorted([OH_m, H2O], key=attrgetter('name')) == [H2O, OH_m]


def test_Species():
    s = Species.from_formula('H2O')
    assert s.phase_idx == 0
    mapping = {'(aq)': 0, '(s)': 1, '(g)': 2}
    assert Species.from_formula('CO2(g)').phase_idx == 3
    assert Species.from_formula('CO2(g)', mapping).phase_idx == 2
    assert Species.from_formula('CO2(aq)', mapping).phase_idx == 0
    assert Species.from_formula('NaCl(s)').phase_idx == 1
    assert Species.from_formula('NaCl(s)', phase_idx=7).phase_idx == 7
    assert Species.from_formula('CO2(aq)', mapping, phase_idx=7).phase_idx == 7


def test_Solute():
    from ..chemistry import Solute
    from ..util.pyutil import ChemPyDeprecationWarning
    with pytest.warns(ChemPyDeprecationWarning):
        w = Solute('H2O')
    assert w.name == 'H2O'


def test_Reaction():
    substances = s_Hp, s_OHm, s_H2O = (
        Substance('H+', composition={0: 1, 1: 1}),
        Substance('OH-', composition={0: -1, 1: 1, 8: 1}),
        Substance('H2O', composition={0: 0, 1: 2, 8: 1}),
    )
    substance_names = Hp, OHm, H2O = [s.name for s in substances]
    substance_dict = {n: s for n, s in zip(substance_names, substances)}
    r1 = Reaction({Hp: 1, OHm: 1}, {H2O: 1})
    assert sum(r1.composition_violation(substance_dict)) == 0
    assert r1.charge_neutrality_violation(substance_dict) == 0

    r2 = Reaction({Hp: 1, OHm: 1}, {H2O: 2})
    assert sum(r2.composition_violation(substance_dict)) != 0
    assert r2.charge_neutrality_violation(substance_dict) == 0

    r3 = Reaction({Hp: 2, OHm: 1}, {H2O: 2})
    assert sum(r3.composition_violation(substance_dict)) != 0
    assert r3.charge_neutrality_violation(substance_dict) != 0


def test_ReactionSystem__as_per_substance_array():
    mol = default_units.mol
    m = default_units.metre
    M = default_units.molar
    rs = ReactionSystem([], [Substance('H2O')])
    c = rs.as_per_substance_array({'H2O': 1*M},
                                  unit=M)
    assert c.dimensionality == M.dimensionality
    assert abs(c[0]/(1000*mol/m**3) - 1) < 1e-16

    c = rs.as_per_substance_array({'H2O': 1})
    with pytest.raises(KeyError):
        c = rs.as_per_substance_array({'H': 1})


def test_ArrheniusRate():
    k = ArrheniusRate(1e10, 42e3)(273.15)
    ref = 1e10 * math.exp(-42e3/(8.3145*273.15))
    assert abs((k - ref)/ref) < 1e-4


def test_ArrheniusRateWithUnits():
    s = default_units.second
    mol = default_units.mol
    J = default_units.joule
    K = default_units.kelvin
    k = ArrheniusRateWithUnits(1e10/s, 42e3 * J/mol)(273.15*K)
    ref = 1e10/s * math.exp(-42e3/(8.3145*273.15))
    assert abs((k - ref)/ref) < 1e-4


def test_Equilibrium__as_reactions():
    s = default_units.second
    M = default_units.molar
    H2O, Hp, OHm = map(Substance, 'H2O H+ OH-'.split())
    eq = Equilibrium({'H2O': 1}, {'H+': 1, 'OH-': 1}, 1e-14)
    rate = 1.31e11/M/s
    fw, bw = eq.as_reactions(kb=rate, units=default_units)
    assert abs((bw.param - rate)/rate) < 1e-15
    assert abs((fw.param / M)/bw.param - 1e-14)/1e-14 < 1e-15


def test_Reaction__from_string():
    r = Reaction.from_string("H2O -> H+ + OH-; 1e-4", 'H2O H+ OH-'.split())
    assert r.reac == {'H2O': 1} and r.prod == {'H+': 1, 'OH-': 1}

    with pytest.raises(ValueError):
        Reaction.from_string("H2O -> H+ + OH-; 1e-4", 'H2O H OH-'.split())



from math import log as ln

from ..debye_huckel import A as A_dh, B as B_dh
from ..debye_huckel import limiting_log_gamma
from ..units import allclose

import quantities as pq


def test_A():
    A20 = A_dh(80.1, 293.15, 998.2071)/ln(10)
    assert abs(A20 - 0.50669) < 1e-5

    A20q = A_dh(80.1, 293.15*pq.K, 998.2071*pq.kg/pq.m**3,
                b0=1*pq.mol/pq.kg, constants=pq.constants)
    assert abs(A20q - 0.50669) < 1e-5


def test_B():
    assert abs(1e-10*B_dh(80.1, 293.15, 998.2071, 1) - 0.3282) < 1e-3

    B20q = B_dh(80.1, 293.15*pq.K, 998.2071*pq.kg/pq.m**3,
                b0=pq.mol/pq.kg, constants=pq.constants, units=pq)
    close = allclose(B20q.simplified, 0.3282/pq.angstrom, rtol=1e-3)
    assert close


def test_limiting_log_gamma():
    A20 = A_dh(80.1, 293.15, 998.2071)/ln(10)
    log_gamma = limiting_log_gamma(0.4, -3, A20)
    assert abs(log_gamma + 2.884130) < 1e-4



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function)

from chempy.chemistry import Solute, Equilibrium
from chempy.equilibria import EqSystem


def get_ammonical_cupric_eqsys():
    NH3_complexes = ['CuNH3/2+', 'Cu(NH3)2/2+', 'Cu(NH3)3/2+', 'Cu(NH3)4/2+',
                     'Cu(NH3)5/2+']
    OH_complexes = ['Cu2(OH)2/2+', 'Cu(OH)3-', 'Cu(OH)4/2-']
    substances = [
        Solute.from_formula(n) for n in [
            'H+', 'OH-', 'NH4+', 'NH3', 'H2O', 'Cu/2+'] +
        NH3_complexes + OH_complexes + ['Cu(OH)2(s)']
    ]

    (Hp, OHm, NH4p, NH3, H2O, Cupp, CuNH31pp, CuNH32pp,
     CuNH33pp, CuNH34pp, CuNH35pp, Cu2OH2pp, CuOH3m,
     CuOH4mm, CuOH2) = [s.name for s in substances]
    substances[-1].precipitate = True
    init_conc = {Hp: 1e-7, OHm: 1e-7, NH4p: 0, NH3: 1.0, Cupp: 1e-2,
                 CuNH31pp: 0, CuNH32pp: 0, CuNH33pp: 0, CuNH34pp: 0,
                 CuNH35pp: 0, H2O: 55.5, Cu2OH2pp: 0, CuOH2: 0, CuOH3m: 0,
                 CuOH4mm: 0}
    H2O_c = init_conc[H2O]
    w_autop = Equilibrium({H2O: 1}, {Hp: 1, OHm: 1}, 10**-14/H2O_c)
    NH4p_pr = Equilibrium({NH4p: 1}, {Hp: 1, NH3: 1}, 10**-9.26)
    CuOH2_s = Equilibrium({CuOH2: 1}, {Cupp: 1, OHm: 2}, 10**-18.8)
    CuOH_B3 = Equilibrium({CuOH2: 1, OHm: 1}, {CuOH3m: 1}, 10**-3.6)
    CuOH_B4 = Equilibrium({CuOH2: 1, OHm: 2}, {CuOH4mm: 1}, 10**-2.7)
    Cu2OH2 = Equilibrium({Cupp: 2, H2O: 2}, {Cu2OH2pp: 1, Hp: 2},
                         10**-10.6 / H2O_c**2)
    CuNH3_B1 = Equilibrium({CuNH31pp: 1}, {Cupp: 1, NH3: 1}, 10**-4.3)
    CuNH3_B2 = Equilibrium({CuNH32pp: 1}, {Cupp: 1, NH3: 2}, 10**-7.9)
    CuNH3_B3 = Equilibrium({CuNH33pp: 1}, {Cupp: 1, NH3: 3}, 10**-10.8)
    CuNH3_B4 = Equilibrium({CuNH34pp: 1}, {Cupp: 1, NH3: 4}, 10**-13.0)
    CuNH3_B5 = Equilibrium({CuNH35pp: 1}, {Cupp: 1, NH3: 5}, 10**-12.4)
    equilibria = (w_autop, NH4p_pr, CuNH3_B1, CuNH3_B2, CuNH3_B3, CuNH3_B4,
                  CuNH3_B5, Cu2OH2, CuOH_B3, CuOH_B4, CuOH2_s)

    new_eqs = CuOH2_s - CuOH_B3, CuOH2_s - CuOH_B4
    skip_subs, skip_eq = (1, 3)
    simpl_subs = substances[:-skip_subs]
    simpl_eq = equilibria[:-skip_eq] + new_eqs
    simpl_c0 = {k.name: init_conc[k.name] for k in substances[:-skip_subs]}
    return EqSystem(simpl_eq, simpl_subs), simpl_c0



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)


from ..henry import Henry, HenryWithUnits
from ..units import allclose, default_units as u


def test_Henry():
    kH_O2 = Henry(1.2e-3, 1800, ref='carpenter_1966')
    assert abs(kH_O2.get_kH_at_T(298.15) - 1.2e-3) < 1e-4
    assert abs(kH_O2.get_c_at_T_and_P(290, 1) - 0.001421892) < 1e-8
    assert abs(kH_O2.get_P_at_T_and_c(310, 1e-3) - 1.05) < 1e-3

    import numpy as np

    kH_H2 = HenryWithUnits(7.8e-4*u.molar/u.atm, 640*u.K, ref='dean_1992')

    assert allclose(kH_H2.get_kH_at_T(
        300*u.K), 7.697430323e-4*u.molar/u.atm)
    kH = kH_H2.get_c_at_T_and_P(
        np.linspace(297.5, 298.65, 3)*u.K, .1*u.bar)
    assert allclose(kH, 7.7e-5*u.molar, rtol=1e-5, atol=1e-6*u.molar)
    kH = kH_H2.get_P_at_T_and_c(
        298.15*u.K, np.linspace(2e-3, 2.1e-3, 3)*u.molar)
    assert allclose(kH, 2.65*u.atm, rtol=1e-5, atol=0.2*u.bar)


def test_HenryWithUnits():
    kH_H2 = HenryWithUnits(7.8e-4*u.molar/u.atm, 640*u.K, ref='dean_1992')
    Hcp = kH_H2(300*u.K)
    assert allclose(Hcp, 7.697430323e-4*u.molar/u.atm)
